"""
    .. autoclass:: BinaryAccuracy()
    .. autoclass:: CategoricalAccuracy(ignore_index=-100)
    .. autoclass:: TopKCategoricalAccuracy(k=5, ignore_index=-100)
    .. autoclass:: MeanSquaredError()
    .. autoclass:: Loss()
    .. autoclass:: Epoch()
"""
import torchbearer
from torchbearer import Metric
from .decorators import default_for_key, running_mean, mean, std, to_dict

import torch

old_super = super


def super(_, obj):
    return old_super(obj.__class__, obj)


@default_for_key('binary_accuracy')
@default_for_key('binary_acc')
@running_mean
@mean
class BinaryAccuracy(Metric):
    """Binary accuracy metric. Uses torch.eq to compare predictions to targets. Decorated with a mean and running_mean.
    Default for key: 'binary_acc'.

    Args:
        pred_key (:class:`torchbearer.StateKey`): The key in state which holds the predicted values
        target_key (:class:`torchbearer.StateKey`): The key in state which holds the target values
        threshold (float): value between 0 and 1 to use as a threshold when binarizing predictions and targets
    """

    def __init__(self, pred_key=torchbearer.Y_PRED, target_key=torchbearer.Y_TRUE, threshold=0.5):
        super(BinaryAccuracy, self).__init__('binary_acc')
        self.pred_key = pred_key
        self.target_key = target_key

        self.threshold = threshold

    def process(self, *args):
        state = args[0]
        y_pred = (state[self.pred_key].float() > self.threshold).long()
        y_true = (state[self.target_key].float() > self.threshold).long()

        return torch.eq(y_pred, y_true).view(-1).float()


@default_for_key('cat_accuracy')
@default_for_key('cat_acc')
@running_mean
@mean
class CategoricalAccuracy(Metric):
    """Categorical accuracy metric. Uses torch.max to determine predictions and compares to targets. Decorated with a
    mean, running_mean and std. Default for key: 'cat_acc'

    Args:
        pred_key (:class:`torchbearer.StateKey`): The key in state which holds the predicted values
        target_key (:class:`torchbearer.StateKey`): The key in state which holds the target values
        ignore_index (int): Specifies a target value that is ignored and does not contribute to the metric output.
            See `<https://pytorch.org/docs/stable/nn.html#crossentropyloss>`_
    """

    def __init__(self, pred_key=torchbearer.Y_PRED, target_key=torchbearer.Y_TRUE, ignore_index=-100):
        super(Metric, self).__init__('acc')
        self.pred_key = pred_key
        self.target_key = target_key

        self.ignore_index = ignore_index

    def process(self, *args):
        state = args[0]
        y_pred = state[self.pred_key]
        y_true = state[self.target_key]
        mask = y_true.eq(self.ignore_index).eq(0)
        y_pred = y_pred[mask]
        y_true = y_true[mask]
        _, y_pred = torch.max(y_pred, 1)
        return (y_pred == y_true).float()


@default_for_key('top_10_accuracy', k=10)
@default_for_key('top_5_accuracy')
@default_for_key('top_10_acc', k=10)
@default_for_key('top_5_acc')
@running_mean
@mean
class TopKCategoricalAccuracy(Metric):
    """Top K Categorical accuracy metric. Uses torch.topk to determine the top k predictions and compares to targets.
    Decorated with a mean, running_mean and std. Default for keys: 'top_5_acc', 'top_10_acc'.

    Args:
        pred_key (:class:`torchbearer.StateKey`): The key in state which holds the predicted values
        target_key (:class:`torchbearer.StateKey`): The key in state which holds the target values
        ignore_index (int): Specifies a target value that is ignored and does not contribute to the metric output.
            See `<https://pytorch.org/docs/stable/nn.html#crossentropyloss>`_
    """

    def __init__(self, pred_key=torchbearer.Y_PRED, target_key=torchbearer.Y_TRUE, k=5, ignore_index=-100):
        super(TopKCategoricalAccuracy, self).__init__('top_' + str(k) + '_acc')
        self.pred_key = pred_key
        self.target_key = target_key

        self.k = k
        self.ignore_index = ignore_index

    def process(self, *args):
        state = args[0]
        y_pred = state[self.pred_key]
        y_true = state[self.target_key]
        mask = y_true.eq(self.ignore_index).eq(0)
        y_pred = y_pred[mask]
        y_true = y_true[mask]

        sorted_indices = torch.topk(y_pred, self.k, dim=1)[1]
        expanded_y = y_true.view(-1, 1).expand(-1, self.k)
        return torch.sum(torch.eq(sorted_indices, expanded_y), dim=1).float()


@default_for_key('mse')
@running_mean
@mean
class MeanSquaredError(Metric):
    """Mean squared error metric. Computes the pixelwise squared error which is then averaged with decorators.
    Decorated with a mean and running_mean. Default for key: 'mse'.

    Args:
        pred_key (:class:`torchbearer.StateKey`): The key in state which holds the predicted values
        target_key (:class:`torchbearer.StateKey`): The key in state which holds the target values
    """

    def __init__(self, pred_key=torchbearer.Y_PRED, target_key=torchbearer.Y_TRUE):
        super(MeanSquaredError, self).__init__('mse')
        self.pred_key = pred_key
        self.target_key = target_key

    def process(self, *args):
        state = args[0]
        y_pred = state[self.pred_key]
        y_true = state[self.target_key]
        return torch.pow(y_pred - y_true.view_as(y_pred), 2).data


@default_for_key('loss')
@running_mean
@mean
class Loss(Metric):
    """Simply returns the 'loss' value from the model state. Decorated with a mean, running_mean and std. Default for
    key: 'loss'.

    State Requirements:
        - :attr:`torchbearer.state.LOSS`: This key should map to the loss for the current batch
    """

    def __init__(self):
        super(Loss, self).__init__('loss')

    def process(self, *args):
        state = args[0]
        return state[torchbearer.LOSS]


@default_for_key('epoch')
@to_dict
class Epoch(Metric):
    """Returns the 'epoch' from the model state. Default for key: 'epoch'.

    State Requirements:
        - :attr:`torchbearer.state.EPOCH`: This key should map to the number of the current epoch
    """

    def __init__(self):
        super(Epoch, self).__init__('epoch')

    def process_final(self, *args):
        state = args[0]
        return self._process(state)

    def process(self, *args):
        state = args[0]
        return self._process(state)

    def _process(self, state):
        return state[torchbearer.EPOCH]



import sys
if sys.version_info[0] < 3:
    import inspect

    def get_default(fcn, arg):
        a = inspect.getargspec(fcn)
        return dict(zip(a.args[-len(a.defaults):], a.defaults))[arg]
else:
    from inspect import signature

    def get_default(fcn, arg):
        return signature(fcn).parameters[arg].default

import functools
import warnings
import itertools

import torch
from torch.utils.data import DataLoader, TensorDataset
from torch.optim import Optimizer

import torchbearer
from torchbearer import cite
from torchbearer import State
from torchbearer.metrics import MetricList
from torchbearer.callbacks import Callback, CallbackList, Tqdm, AggregatePredictions


bibtex = """
@article{2018torchbearer,
  title={Torchbearer: A Model Fitting Library for PyTorch},
  author={Harris, Ethan and Painter, Matthew and Hare, Jonathon},
  journal={arXiv preprint arXiv:1809.03363},
  year={2018}
}
"""


class MockOptimizer(Optimizer):
    """The Mock Optimizer will be used inplace of an optimizer in the event that none is passed to the Trial class.
    """
    def __init__(self):
        super(MockOptimizer, self).__init__([torch.zeros(1)], [])

    def add_param_group(self, param_group):
        pass  # Do Nothing

    def load_state_dict(self, state_dict):
        pass  # Do Nothing

    def state_dict(self):
        return {}  # Return Empty

    def step(self, closure=None):
        pass  # Do Nothing

    def zero_grad(self):
        pass  # Do Nothing


class CallbackListInjection(CallbackList):
    """This class allows for an callback to be injected into a callback list, without masking the methods available for
    mutating the list. In this way, callbacks (such as printers) can be injected seamlessly into the methods of the
    trial class.

    Args:
        callback: The callback to inject
        callback_list (:class:`.CallbackList`): The underlying callback list
    """
    def __init__(self, callback, callback_list):
        super(CallbackListInjection, self).__init__([])

        self.callback = callback
        self.callback_list = callback_list

    def state_dict(self):
        return self.callback_list.state_dict()

    def load_state_dict(self, state_dict):
        self.callback_list.load_state_dict(state_dict)
        return self

    def __iter__(self):
        return self.callback_list.__iter__()

    def __copy__(self):
        return self.callback_list.copy()

    def copy(self):
        return self.__copy__()

    def append(self, callback_list):
        self.callback_list.append(callback_list)

    def _for_list(self, function):
        function(self.callback)  # Call injected callback BEFORE the callback list
        function(self.callback_list)


def inject_printer(validation_label_letter='v'):
    """The inject printer decorator is used to inject the appropriate printer callback, according to the verbosity level.

    Args:
        validation_label_letter (str): The validation label letter to use

    Returns:
        A decorator
    """
    def decorator(func):
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            verbose = kwargs['verbose'] if 'verbose' in kwargs else get_default(func, 'verbose')  # Populate default value
            verbose = self.verbose if verbose == -1 else verbose

            printer = get_printer(verbose=verbose, validation_label_letter=validation_label_letter)

            callback_list_old = self.state[torchbearer.CALLBACK_LIST]

            self.state[torchbearer.CALLBACK_LIST] = CallbackListInjection(printer, callback_list_old)

            res = func(self, *args, **kwargs)

            self.state[torchbearer.CALLBACK_LIST] = callback_list_old
            return res
        return wrapper
    return decorator


def get_printer(verbose, validation_label_letter):
    if verbose >= 2:
        printer = Tqdm(validation_label_letter=validation_label_letter)
    elif verbose >= 1:
        printer = Tqdm(validation_label_letter=validation_label_letter, on_epoch=True)
    else:
        printer = Callback()

    return printer


def deep_to(batch, device, dtype):
    """ Static method to call :func:`to` on tensors or tuples. All items in tuple will have :func:`deep_to` called

    Args:
        batch (tuple / list / :class:`torch.Tensor`): The mini-batch which requires a :func:`to` call
        device (:class:`torch.device`): The desired device of the batch
        dtype (:class:`torch.dtype`): The desired datatype of the batch

    Returns:
        tuple / list / :class:`torch.Tensor`: The moved or casted batch
    """
    is_tuple = isinstance(batch, tuple)

    if isinstance(batch, list) or isinstance(batch, tuple):
        batch = list(batch)
        for i in range(len(batch)):
            batch[i] = deep_to(batch[i], device, dtype)
        batch = tuple(batch) if is_tuple else batch
    elif isinstance(batch, dict):
        for key in batch:
            batch[key] = deep_to(batch[key], device, dtype)
    else:
        if batch.dtype.is_floating_point:
            batch = batch.to(device, dtype)
        else:
            batch = batch.to(device)

    return batch


def load_batch_infinite(loader):
    """ Wraps a batch loader and refreshes the iterator once it has been completed.

    Args:
        loader: batch loader to wrap
    """

    def call(state):
        try:
            loader(state)
        except StopIteration:
            state[torchbearer.ITERATOR] = iter(state[torchbearer.GENERATOR])
            loader(state)

    return call


def load_batch_standard(state):
    """ Load a standard (input data, target) tuple mini-batch from iterator into state

    Args:
        state (dict): The current state dict of the :class:`Trial`.
    """
    state[torchbearer.X], state[torchbearer.Y_TRUE] = deep_to(next(state[torchbearer.ITERATOR]),
                                                              state[torchbearer.DEVICE],
                                                              state[torchbearer.DATA_TYPE])


def load_batch_none(state):
    """ Load a none (none, none) tuple mini-batch into state

    Args:
        state (dict): The current state dict of the :class:`Trial`.
    """
    state[torchbearer.X], state[torchbearer.Y_TRUE] = None, None


def load_batch_predict(state):
    """ Load a prediction (input data, target) or (input data) mini-batch from iterator into state

    Args:
        state (dict): The current state dict of the :class:`Trial`.
    """
    data = deep_to(next(state[torchbearer.ITERATOR]), state[torchbearer.DEVICE], state[torchbearer.DATA_TYPE])
    if isinstance(data, list) or isinstance(data, tuple):
        state[torchbearer.X], state[torchbearer.Y_TRUE] = data
    else:
        state[torchbearer.X] = data
        
        
class Sampler:
    """
    Sampler wraps a batch loader function and executes it when :meth:`Sampler.sample` is called

    Args:
        batch_loader (func): The batch loader to execute
    """
    def __init__(self, batch_loader):
        self.batch_loader = batch_loader

    def sample(self, state):
        self.batch_loader(state)


def inject_sampler(data_key, predict=False):
    """ Decorator to inject a :class:`Sampler` into state[torchbearer.SAMPLER] along with the specified \
        generator into state[torchbearer.GENERATOR] and number of steps into state[torchbearer.STEPS]

    Args:
        data_key (:class:`.StateKey`): Key for the data to inject
        predict (bool): If true, the prediction batch loader is used, if false the standard data loader is used

    Returns:
        The decorator
    """
    def decorator(func):
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            key = kwargs['data_key'] if 'data_key' in kwargs else data_key  # Populate default value
            generator, steps = self.state[key] if self.state[key] is not None else (None, None)

            if generator is None:
                loader = load_batch_none
            elif predict:
                loader = load_batch_predict
            else:
                loader = load_batch_standard

            if generator is not None and steps is not None:
                over_steps = steps > len(generator)
                inf_steps = steps == -1
                inf_train_loader = key == torchbearer.TRAIN_DATA and self.state[torchbearer.INF_TRAIN_LOADING]
                if over_steps or inf_steps or inf_train_loader: # Want iterator to refresh at end
                    if steps == -1: warnings.warn("Trial is set to run indefinitely. "
                                              "Make sure you have some method to terminate safely.")
                    loader = load_batch_infinite(loader)

                if inf_train_loader and not hasattr(generator, 'inf'): # First run and infinite iterator
                    generator.inf = True
                    generator.tb_iter = iter(generator)

            self.state[torchbearer.DATA] = key
            self.state[torchbearer.SAMPLER] = Sampler(loader)
            self.state[torchbearer.GENERATOR] = generator
            self.state[torchbearer.STEPS] = steps

            res = func(self, *args, **kwargs)

            return res
        return wrapper
    return decorator


def inject_callback(callback):
    """ Decorator to inject a callback into the callback list and remove the callback after the decorated function has executed

    Args:
        callback (:class:`.Callback`): Callback to be injected

    Returns:
        The decorator
    """
    def decorator(func):
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            callback_list_old = self.state[torchbearer.CALLBACK_LIST]

            self.state[torchbearer.CALLBACK_LIST] = CallbackListInjection(callback, callback_list_old)

            res = func(self, *args, **kwargs)

            self.state[torchbearer.CALLBACK_LIST] = callback_list_old
            return res
        return wrapper
    return decorator


def fluent(func):
    """Decorator for class methods which forces return of self.
    """
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        func(self, *args, **kwargs)
        return self
    return wrapper


def update_device_and_dtype(state, *args, **kwargs):
    """Function get data type and device values from the args / kwargs and update state.

    Args:
        state (:class:`.State`): The dict to update
        args: Arguments to the :func:`Trial.to` function
        kwargs: Keyword arguments to the :func:`Trial.to` function

    Returns:
        device, dtype pair
    """
    for key, _ in kwargs.items():
        if key == str(torchbearer.DATA_TYPE):
            state[torchbearer.DATA_TYPE] = kwargs['dtype']
        elif str(torchbearer.DEVICE) in kwargs:
            state[torchbearer.DEVICE] = kwargs['device']

    for arg in args:
        if isinstance(arg, torch.dtype):
            state[torchbearer.DATA_TYPE] = arg
        else:
            state[torchbearer.DEVICE] = arg

    return state


@cite(bibtex)
class Trial(object):
    """
    The trial class contains all of the required hyper-parameters for model running in torchbearer and presents an
    API for model fitting, evaluating and predicting.

    Args:
        model (:class:`torch.nn.Module`): The base pytorch model
        optimizer (:class:`torch.optim.Optimizer`): The optimizer used for pytorch model weight updates
        criterion (func / None): The final loss criterion that provides a loss value to the optimizer
        metrics (list): The list of :class:`torchbearer.Metric <.Metric>` instances to process during fitting
        callbacks (list): The list of :class:`torchbearer.Callback <.Callback>` instances to call during fitting
        verbose (int): Global verbosity .If 2: use tqdm on batch, If 1: use tqdm on epoch, If 0: display no training
            progress
    """
    def __init__(self, model, optimizer=None, criterion=None, metrics=[], callbacks=[], verbose=2):
        if criterion is None:
            def criterion(_, __):
                return torch.zeros(1, device=self.state[torchbearer.DEVICE], dtype=self.state[torchbearer.DATA_TYPE], requires_grad=True)

        self.verbose = verbose

        self.state = State()
        self.state.update({
            torchbearer.MODEL: model,
            torchbearer.CRITERION: criterion,
            torchbearer.OPTIMIZER: optimizer if optimizer is not None else MockOptimizer(),
            torchbearer.METRIC_LIST: MetricList(metrics),
            torchbearer.CALLBACK_LIST: CallbackList(callbacks),
            torchbearer.DEVICE: 'cpu',
            torchbearer.DATA_TYPE: torch.float32,
            torchbearer.SELF: self,
            torchbearer.HISTORY: [],
            torchbearer.BACKWARD_ARGS: {},
            torchbearer.TRAIN_GENERATOR: None,
            torchbearer.VALIDATION_GENERATOR: None,
            torchbearer.TEST_GENERATOR: None,
            torchbearer.TRAIN_STEPS: None,
            torchbearer.VALIDATION_STEPS: None,
            torchbearer.TEST_STEPS: None,
            torchbearer.TRAIN_DATA: None,
            torchbearer.VALIDATION_DATA: None,
            torchbearer.TEST_DATA: None,
            torchbearer.INF_TRAIN_LOADING: False,
        })

    def __str__(self):
        def state_string(name, state_key):
            import math
            N = (50-len(name))/2
            res = "-" * int(math.floor(N)) + " " + name.upper() + " " + "-" * int(math.ceil(N))
            res = res + "-" if len(res) < 52 else res
            return res + "\n" + str(self.state[state_key]) + "\n\n"

        optim_str = state_string('Optimzer', torchbearer.OPTIMIZER)
        crit_str = state_string("Criterion", torchbearer.CRITERION)
        metrics_str = state_string("Metrics", torchbearer.METRIC_LIST)
        callbacks_str = state_string("Callbacks", torchbearer.CALLBACK_LIST)
        model_str = state_string("Model", torchbearer.MODEL)

        return optim_str + crit_str + metrics_str + callbacks_str + model_str

    def __repr__(self):
        return str(self)

    @fluent
    def for_train_steps(self, steps):
        """Run this trial for the given number of training steps. Note that the generator will output (None, None) if it
        has not been set. Useful for differentiable programming. Returns self so that methods can be chained for
        convenience. If steps is larger than dataset size then loader will be refreshed like if it was a new epoch. If
        steps is -1 then loader will be refreshed until stopped by STOP_TRAINING flag or similar.

        Args:
            steps (int): The number of training steps per epoch to run.

        Returns:
            :class:`.Trial`: self
        """
        if not isinstance(steps, int):
            warnings.warn("Number of training steps is not an int, casting to int")
            steps = int(steps)
        self.state[torchbearer.TRAIN_STEPS] = steps
        self.state[torchbearer.TRAIN_DATA] = (self.state[torchbearer.TRAIN_GENERATOR], self.state[torchbearer.TRAIN_STEPS])

    @fluent
    def with_train_generator(self, generator, steps=None):
        """Use this trial with the given train generator. Returns self so that methods can be chained for convenience.

        Args:
            generator: The train data generator to use during calls to :meth:`.run`
            steps (int): The number of steps per epoch to take when using this generator.

        Returns:
            :class:`.Trial`: self
        """
        self.state[torchbearer.TRAIN_GENERATOR] = generator
        steps = len(generator) if steps is None else steps
        self.for_train_steps(steps)

    @fluent
    def with_train_data(self, x, y, batch_size=1, shuffle=True, num_workers=1, steps=None):
        """Use this trial with the given train data. Returns self so that methods can be chained for convenience.

        Args:
            x (:class:`torch.Tensor`): The train x data to use during calls to :meth:`.run`
            y (:class:`torch.Tensor`): The train labels to use during calls to :meth:`.run`
            batch_size (int): The size of each batch to sample from the data
            shuffle (bool): If True, then data will be shuffled each epoch
            num_workers (int): Number of worker threads to use in the data loader
            steps (int): The number of steps per epoch to take when using this data

        Returns:
            :class:`.Trial`: self
        """
        dataset = TensorDataset(x, y)
        dataloader = DataLoader(dataset, batch_size, shuffle=shuffle, num_workers=num_workers)
        self.with_train_generator(dataloader, steps=steps)

    @fluent
    def for_val_steps(self, steps):
        """Run this trial for the given number of validation steps. Note that the generator will output (None, None) if
        it has not been set. Useful for differentiable programming. Returns self so that methods can be chained for
        convenience. If steps larger than dataset size then loader will be refreshed like if it was a new epoch. If
        steps -1 then loader will be refreshed until stopped by STOP_TRAINING flag or similar.

        Args:
            steps (int): The number of validation steps per epoch to run

        Returns:
            :class:`.Trial`: self
        """
        if not isinstance(steps, int):
            warnings.warn("Number of validation steps is not an int, casting to int")
            steps = int(steps)
        self.state[torchbearer.VALIDATION_STEPS] = steps
        self.state[torchbearer.VALIDATION_DATA] = (self.state[torchbearer.VALIDATION_GENERATOR], self.state[torchbearer.VALIDATION_STEPS])

    @fluent
    def with_val_generator(self, generator, steps=None):
        """Use this trial with the given validation generator. Returns self so that methods can be chained for
        convenience.

        Args:
            generator: The validation data generator to use during calls to :meth:`.run` and :meth:`.evaluate`
            steps (int): The number of steps per epoch to take when using this generator

        Returns:
            :class:`.Trial`: self
        """
        self.state[torchbearer.VALIDATION_GENERATOR] = generator
        steps = len(generator) if steps is None else steps
        self.for_val_steps(steps)

    @fluent
    def with_val_data(self, x, y, batch_size=1, shuffle=True, num_workers=1, steps=None):
        """Use this trial with the given validation data. Returns self so that methods can be chained for convenience.

        Args:
            x (:class:`torch.Tensor`): The validation x data to use during calls to :meth:`.run` and :meth:`.evaluate`
            y (:class:`torch.Tensor`): The validation labels to use during calls to :meth:`.run` and :meth:`.evaluate`
            batch_size (int): The size of each batch to sample from the data
            shuffle (bool): If True, then data will be shuffled each epoch
            num_workers (int): Number of worker threads to use in the data loader
            steps (int): The number of steps per epoch to take when using this data

        Returns:
            :class:`.Trial`: self
        """
        dataset = TensorDataset(x, y)
        dataloader = DataLoader(dataset, batch_size, shuffle=shuffle, num_workers=num_workers)
        self.with_val_generator(dataloader, steps=steps)

    @fluent
    def for_test_steps(self, steps):
        """Run this trial for the given number of test steps. Note that the generator will output (None, None) if
        it has not been set. Useful for differentiable programming. Returns self so that methods can be chained for
        convenience. If steps larger than dataset size then loader will be refreshed like if it was a new epoch. If
        steps -1 then loader will be refreshed until stopped by STOP_TRAINING flag or similar.

        Args:
            steps (int): The number of test steps per epoch to run (when using :meth:`.predict`)

        Returns:
            :class:`.Trial`: self
        """
        if not isinstance(steps, int):
            warnings.warn("Number of test steps is not an int, casting to int")
            steps = int(steps)
        self.state[torchbearer.TEST_STEPS] = steps
        self.state[torchbearer.TEST_DATA] = (self.state[torchbearer.TEST_GENERATOR], self.state[torchbearer.TEST_STEPS])

    @fluent
    def with_test_generator(self, generator, steps=None):
        """Use this trial with the given test generator. Returns self so that methods can be chained for convenience.

        Args:
            generator: The test data generator to use during calls to :meth:`.predict`
            steps (int): The number of steps per epoch to take when using this generator

        Returns:
            :class:`.Trial`: self
        """
        self.state[torchbearer.TEST_GENERATOR] = generator
        steps = len(generator) if steps is None else steps
        self.for_test_steps(steps)

    @fluent
    def with_test_data(self, x, batch_size=1, num_workers=1, steps=None):
        """Use this trial with the given test data. Returns self so that methods can be chained for convenience.

        Args:
            x (:class:`torch.Tensor`): The test x data to use during calls to :meth:`.predict`
            batch_size (int): The size of each batch to sample from the data
            num_workers (int): Number of worker threads to use in the data loader
            steps (int): The number of steps per epoch to take when using this data

        Returns:
            :class:`.Trial`: self
        """
        dataset = TensorDataset(x)
        dataloader = DataLoader(dataset, batch_size, num_workers=num_workers)
        self.with_test_generator(dataloader, steps=steps)

    @fluent
    def for_steps(self, train_steps=None, val_steps=None, test_steps=None):
        """Use this trial for the given number of train, val and test steps. Returns self so that methods can be chained
        for convenience. If steps larger than dataset size then loader will be refreshed like if it was a new epoch. If
        steps -1 then loader will be refreshed until stopped by STOP_TRAINING flag or similar.

        Args:
            train_steps (int): The number of training steps per epoch to run
            val_steps (int): The number of validation steps per epoch to run
            test_steps (int): The number of test steps per epoch to run (when using :meth:`.predict`)

        Returns:
            :class:`.Trial`: self
        """
        if train_steps is not None:
            self.for_train_steps(train_steps)
        if val_steps is not None:
            self.for_val_steps(val_steps)
        if test_steps is not None:
            self.for_test_steps(test_steps)

    @fluent
    def with_generators(self, train_generator=None, val_generator=None, test_generator=None, train_steps=None, val_steps=None, test_steps=None):
        """Use this trial with the given generators. Returns self so that methods can be chained for convenience.

        Args:
            train_generator: The training data generator to use during calls to :meth:`.run`
            val_generator: The validation data generator to use during calls to :meth:`.run` and :meth:`.evaluate`
            test_generator: The testing data generator to use during calls to :meth:`.predict`
            train_steps (int): The number of steps per epoch to take when using the training generator
            val_steps (int): The number of steps per epoch to take when using the validation generator
            test_steps (int): The number of steps per epoch to take when using the testing generator

        Returns:
            :class:`.Trial`: self
        """
        if train_generator is not None:
            self.with_train_generator(train_generator, train_steps)
        if val_generator is not None:
            self.with_val_generator(val_generator, val_steps)
        if test_generator is not None:
            self.with_test_generator(test_generator, test_steps)

    @fluent
    def for_inf_train_steps(self):
        """
        Use this trial with an infinite number of training steps (until stopped via STOP_TRAINING flag or similar). Returns self so that methods can be chained for convenience.

        Returns:
            :class:`.Trial`: self
        """
        self.for_train_steps(-1)

    @fluent
    def for_inf_val_steps(self):
        """
        Use this trial with an infinite number of validation steps (until stopped via STOP_TRAINING flag or similar). Returns self so that methods can be chained for convenience.

        Returns:
            :class:`.Trial`: self
        """
        self.for_val_steps(-1)

    @fluent
    def for_inf_test_steps(self):
        """
        Use this trial with an infinite number of test steps (until stopped via STOP_TRAINING flag or similar). Returns self so that methods can be chained for convenience.

        Returns:
            :class:`.Trial`: self
        """
        self.for_test_steps(-1)

    @fluent
    def for_inf_steps(self, train=True, val=True, test=True):
        """
        Use this trail with infinite steps. Returns self so that methods can be chained for convenience.

        Args:
            train (bool): Use an infinite number of training steps
            val (bool): Use an infinite number of validation steps
            test (bool): Use an infinite number of test steps

        Returns:
            :class:`.Trial`: self
        """
        if train: self.for_inf_train_steps()
        if val: self.for_inf_val_steps()
        if test: self.for_inf_test_steps()

    @fluent
    def with_inf_train_loader(self):
        """
        Use this trial with a training iterator that refreshes when it finishes instead of each epoch. This allows for setting training steps less than the size of the generator and model will still be trained on all training samples if enough "epochs" are run.

        Returns:
            :class:`.Trial`: self:
        """
        self.state[torchbearer.INF_TRAIN_LOADING] = True

    @inject_printer()
    def run(self, epochs=1, verbose=-1):
        r"""Run this trial for the given number of epochs, starting from the last trained epoch.

        Args:
            epochs (int, optional): The number of epochs to run for
            verbose (int, optional): If 2: use tqdm on batch, If 1: use tqdm on epoch, If 0: display no training
            progress, If -1: Automatic

        State Requirements:
            - :attr:`torchbearer.state.MODEL`: Model should be callable and not none, set on Trial init

        Returns:
            list: The model history (list of tuple of steps summary and epoch metric dicts)
        """
        state = State()
        state.update({
            torchbearer.MAX_EPOCHS: epochs,
            torchbearer.STOP_TRAINING: False,
        })

        state.update(self.state)  # TODO: Swap this for something which makes `self.state` still mutable

        if state[torchbearer.MODEL] is None or not callable(state[torchbearer.MODEL]):
            warnings.warn('The Model is None or not callable which may cause issues if not deliberate')
            state[torchbearer.MODEL] = lambda *args, **kwargs: None

        if state[torchbearer.TRAIN_GENERATOR] is not None \
                or state[torchbearer.TRAIN_STEPS] is not None \
                or state[torchbearer.VALIDATION_GENERATOR] is not None \
                or state[torchbearer.VALIDATION_STEPS] is not None:

            state[torchbearer.CALLBACK_LIST].on_start(state)

            for state[torchbearer.EPOCH] in range(len(state[torchbearer.HISTORY]), state[torchbearer.MAX_EPOCHS]):
                state[torchbearer.CALLBACK_LIST].on_start_epoch(state)

                final_metrics = self._fit_pass(state)[torchbearer.METRICS]

                if state[torchbearer.STOP_TRAINING]:
                    break

                final_metrics.update(self._validation_pass(state))
                state[torchbearer.METRICS] = final_metrics
                state[torchbearer.CALLBACK_LIST].on_end_epoch(state)
                steps_summary = (state[torchbearer.TRAIN_STEPS], state[torchbearer.VALIDATION_STEPS])
                self.state[torchbearer.HISTORY].append((steps_summary, state[torchbearer.METRICS]))
                state[torchbearer.CALLBACK_LIST].on_checkpoint(state)

                if state[torchbearer.STOP_TRAINING]:
                    break

            state[torchbearer.CALLBACK_LIST].on_end(state)

        return self.state[torchbearer.HISTORY]

    @staticmethod
    def _new_iter(generator):
        if generator is None:
            return None
        if hasattr(generator, 'inf') and generator.inf: # Inf train loader deals with the iterator itself
            return generator.tb_iter
        else:
            return iter(generator)

    @inject_sampler(torchbearer.TRAIN_DATA)
    def _fit_pass(self, state):
        state.update(self.state)  # TODO: Hack to make injection work, should be removed if `self.state` is mutable
        self.train()

        state[torchbearer.ITERATOR] = Trial._new_iter(state[torchbearer.GENERATOR])

        state[torchbearer.METRIC_LIST].reset(state)
        state[torchbearer.METRICS] = {}

        state[torchbearer.CALLBACK_LIST].on_start_training(state)
        for state[torchbearer.BATCH] in (range(state[torchbearer.STEPS]) if state[torchbearer.STEPS] != -1 else itertools.count()) :
            state[torchbearer.SAMPLER].sample(state)
            state[torchbearer.CALLBACK_LIST].on_sample(state)

            def closure():
                # Zero grads
                state[torchbearer.OPTIMIZER].zero_grad()

                # Forward Pass
                try:
                    state[torchbearer.Y_PRED] = state[torchbearer.MODEL](state[torchbearer.X], state=state)
                except TypeError:
                    state[torchbearer.Y_PRED] = state[torchbearer.MODEL](state[torchbearer.X])

                state[torchbearer.CALLBACK_LIST].on_forward(state)

                # Loss Calculation
                state[torchbearer.LOSS] = state[torchbearer.CRITERION](state[torchbearer.Y_PRED], state[torchbearer.Y_TRUE])

                state[torchbearer.CALLBACK_LIST].on_criterion(state)

                # Backwards pass
                state[torchbearer.LOSS].backward(**state[torchbearer.BACKWARD_ARGS])
                state[torchbearer.CALLBACK_LIST].on_backward(state)

            # Update parameters
            state[torchbearer.OPTIMIZER].step(closure)
            state[torchbearer.METRICS] = state[torchbearer.METRIC_LIST].process(state)
            state[torchbearer.CALLBACK_LIST].on_step_training(state)

            if state[torchbearer.STOP_TRAINING]:
                break

        state[torchbearer.METRICS].update(state[torchbearer.METRIC_LIST].process_final(state))

        state[torchbearer.CALLBACK_LIST].on_end_training(state)
        return state

    def _test_pass(self, state):
        with torch.no_grad():
            state[torchbearer.ITERATOR] = iter(state[torchbearer.GENERATOR]) if state[torchbearer.GENERATOR] is not None else None  # TODO: Inject this?

            state[torchbearer.METRIC_LIST].reset(state)
            state[torchbearer.METRICS] = {}

            state[torchbearer.CALLBACK_LIST].on_start_validation(state)

            for state[torchbearer.BATCH] in range(state[torchbearer.STEPS]):
                state[torchbearer.SAMPLER].sample(state)
                state[torchbearer.CALLBACK_LIST].on_sample_validation(state)

                # Forward Pass
                try:
                    state[torchbearer.Y_PRED] = state[torchbearer.MODEL](state[torchbearer.X], state=state)
                except TypeError:
                    state[torchbearer.Y_PRED] = state[torchbearer.MODEL](state[torchbearer.X])

                state[torchbearer.CALLBACK_LIST].on_forward_validation(state)

                # Loss and metrics
                if torchbearer.Y_TRUE in state:
                    state[torchbearer.LOSS] = state[torchbearer.CRITERION](state[torchbearer.Y_PRED],
                                                                           state[torchbearer.Y_TRUE])
                    state[torchbearer.CALLBACK_LIST].on_criterion_validation(state)
                    state[torchbearer.METRICS] = state[torchbearer.METRIC_LIST].process(state)

                state[torchbearer.CALLBACK_LIST].on_step_validation(state)
                if state[torchbearer.STOP_TRAINING]:
                    break

            if torchbearer.Y_TRUE in state:
                state[torchbearer.METRICS].update(state[torchbearer.METRIC_LIST].process_final(state))
            state[torchbearer.CALLBACK_LIST].on_end_validation(state)
        return state

    @inject_sampler(torchbearer.VALIDATION_DATA)
    def _validation_pass(self, state):
        state.update(self.state)  # TODO: Hack to make injection work, should be removed if `self.state` is mutable

        if state[torchbearer.VALIDATION_GENERATOR] is not None or state[torchbearer.VALIDATION_STEPS] is not None:
            self.eval()

            self._test_pass(state)
        return state[torchbearer.METRICS]

    @inject_sampler(torchbearer.VALIDATION_DATA)
    @inject_printer(validation_label_letter='e')
    def evaluate(self, verbose=-1, data_key=None):  # Note: kwargs appear unused but are inspected in inject_sampler
        """Evaluate this trial on the validation data.

        Args:
            verbose (int): If 2: use tqdm on batch, If 1: use tqdm on epoch, If 0: display no training progress, If -1: Automatic
            data_key (:class:`.StateKey`): Optional key for the data to evaluate on. Default: torchbearer.VALIDATION_DATA

        Returns:
            dict: The final metric values
        """
        state = State()
        state.update({
            torchbearer.MAX_EPOCHS: 1,
            torchbearer.EPOCH: 0,
            torchbearer.STOP_TRAINING: False
        })
        state.update(self.state)  # TODO: Hack to make injection work, should be removed if `self.state` is mutable

        if state[torchbearer.GENERATOR] is not None or state[torchbearer.STEPS] is not None:
            state[torchbearer.CALLBACK_LIST].on_start(state)
            state[torchbearer.CALLBACK_LIST].on_start_epoch(state)

            self.eval()
            state = self._test_pass(state)

            state[torchbearer.CALLBACK_LIST].on_end_epoch(state)

            if len(self.state[torchbearer.HISTORY]) != 0:
                self.state[torchbearer.HISTORY][-1][1].update(state[torchbearer.METRICS])

            state[torchbearer.CALLBACK_LIST].on_end(state)
            return state[torchbearer.METRICS]
        return {}

    @inject_callback(AggregatePredictions())
    @inject_sampler(torchbearer.TEST_DATA, predict=True)
    @inject_printer(validation_label_letter='p')
    def predict(self, verbose=-1, data_key=None):  # Note: kwargs appear unused but are inspected in inject_sampler
        """Determine predictions for this trial on the test data.

        Args:
            verbose (int): If 2: use tqdm on batch, If 1: use tqdm on epoch, If 0: display no training progress, If -1: Automatic
            data_key (:class:`.StateKey`): Optional key for the data to predict on. Default: torchbearer.TEST_DATA

        Returns:
            list: Model outputs as a list
        """
        state = {
            torchbearer.MAX_EPOCHS: 1,
            torchbearer.EPOCH: 0,
            torchbearer.STOP_TRAINING: False
        }

        state.update(self.state)  # TODO: Hack to make injection work, should be removed if `self.state` is mutable

        if state[torchbearer.GENERATOR] is not None or state[torchbearer.STEPS] is not None:
            state[torchbearer.CALLBACK_LIST].on_start(state)
            state[torchbearer.CALLBACK_LIST].on_start_epoch(state)

            self.eval()
            res = self._test_pass(state)[torchbearer.FINAL_PREDICTIONS]

            state[torchbearer.CALLBACK_LIST].on_end_epoch(state)
            state[torchbearer.CALLBACK_LIST].on_end(state)
            return res
        return []

    @fluent
    def replay(self, callbacks=[], verbose=2, one_batch=False):  # TODO: Should we track if testing passes have happened?
        """ Replay the fit passes stored in history with given callbacks, useful when reloading a saved Trial. Note that only progress and metric information is populated in state during a replay.

        Args:
            callbacks (list): List of callbacks to be run during the replay
            verbose (int): If 2: use tqdm on batch, If 1: use tqdm on epoch, If 0: display no training progress
            one_batch (bool): If True, only one batch per epoch is replayed. If False, all batches are replayed

        Returns:
            :class:`.Trial`: self
        """
        history = self.state[torchbearer.HISTORY]
        callbacks.append(get_printer(verbose=verbose, validation_label_letter='v'))
        callbacks = CallbackList(callbacks)

        state = State()
        state.update(self.state)
        state[torchbearer.STOP_TRAINING] = False
        state[torchbearer.MAX_EPOCHS] = len(history)

        callbacks.on_start(state)
        for i in range(len(history)):
            state[torchbearer.EPOCH] = i
            if not one_batch:
                state[torchbearer.TRAIN_STEPS], state[torchbearer.VALIDATION_STEPS] = history[i][0]
            else:
                state[torchbearer.TRAIN_STEPS], state[torchbearer.VALIDATION_STEPS] = 1, 1
            state[torchbearer.METRICS] = history[i][1]

            self._replay_pass(state, callbacks)
        callbacks.on_end(state)

    @fluent
    def _replay_pass(self, state, callback_list):
        callback_list.on_start_epoch(state)
        all_metrics = state[torchbearer.METRICS]

        # Training pass
        state[torchbearer.STEPS] = state[torchbearer.TRAIN_STEPS]
        state[torchbearer.METRICS] = {key: all_metrics[key] for key in all_metrics.keys() if "val_" not in key}
        callback_list.on_start_training(state)
        for state[torchbearer.BATCH] in range(state[torchbearer.STEPS]):
            callback_list.on_sample(state)
            callback_list.on_forward(state)
            callback_list.on_criterion(state)
            callback_list.on_backward(state)
            callback_list.on_step_training(state)
            if state[torchbearer.STOP_TRAINING]:
                break
        callback_list.on_end_training(state)

        # Validation pass
        if not state[torchbearer.STOP_TRAINING]:
            state[torchbearer.STEPS] = state[torchbearer.VALIDATION_STEPS]
            state[torchbearer.METRICS] = {key: all_metrics[key] for key in all_metrics.keys() if "val_" in key}
            callback_list.on_start_validation(state)
            for state[torchbearer.BATCH] in range(state[torchbearer.STEPS]):
                callback_list.on_sample_validation(state)
                callback_list.on_forward_validation(state)
                callback_list.on_criterion_validation(state)
                callback_list.on_step_validation(state)
                if state[torchbearer.STOP_TRAINING]:
                    break
            callback_list.on_end_validation(state)

        state[torchbearer.METRICS] = all_metrics
        callback_list.on_end_epoch(state)

    @fluent
    def train(self):
        """Set model and metrics to training mode.

        Returns:
            :class:`.Trial`: self
        """
        self.state[torchbearer.MODEL].train()
        self.state[torchbearer.METRIC_LIST].train()

    @fluent
    def eval(self):
        """Set model and metrics to evaluation mode

        Returns:
            :class:`.Trial`: self
        """
        self.state[torchbearer.MODEL].eval()
        if torchbearer.DATA in self.state:
            self.state[torchbearer.METRIC_LIST].eval(data_key=self.state[torchbearer.DATA])
        else:
            self.state[torchbearer.METRIC_LIST].eval()

    @fluent
    def to(self, *args, **kwargs):
        """ Moves and/or casts the parameters and buffers.

        Args:
            args: See: `torch.nn.Module.to <https://pytorch.org/docs/stable/nn.html?highlight=#torch.nn.Module.to>`_
            kwargs: See: `torch.nn.Module.to <https://pytorch.org/docs/stable/nn.html?highlight=#torch.nn.Module.to>`_

        Returns:
            :class:`.Trial`: self
        """
        self.state[torchbearer.MODEL].to(*args, **kwargs)

        for state in self.state[torchbearer.OPTIMIZER].state.values():
            for k, v in state.items():
                if torch.is_tensor(v):
                    state[k] = v.to(*args, **kwargs)

        self.state = update_device_and_dtype(self.state, *args, **kwargs)

    @fluent
    def cuda(self, device=None):
        """ Moves all model parameters and buffers to the GPU.

        Args:
            device (int): if specified, all parameters will be copied to that device

        Returns:
            :class:`.Trial`: self
        """
        if device is None:
            device = torch.cuda.current_device()
        self.to('cuda:' + str(device))

    @fluent
    def cpu(self):
        """ Moves all model parameters and buffers to the CPU.

        Returns:
            :class:`.Trial`: self
        """
        self.to('cpu')

    def state_dict(self, **kwargs):
        """Get a dict containing the model and optimizer states, as well as the model history.

        Args:
            kwargs: See: `torch.nn.Module.state_dict <https://pytorch.org/docs/stable/nn.html?highlight=#torch.nn.Module.state_dict>`_

        Returns:
            dict: A dict containing parameters and persistent buffers.
        """
        state_dict = {
            torchbearer.VERSION: torchbearer.__version__.replace('.dev', ''),
            torchbearer.MODEL: self.state[torchbearer.MODEL].state_dict(**kwargs),
            torchbearer.OPTIMIZER: self.state[torchbearer.OPTIMIZER].state_dict(),
            torchbearer.HISTORY: self.state[torchbearer.HISTORY],
            torchbearer.CALLBACK_LIST: self.state[torchbearer.CALLBACK_LIST].state_dict()
        }
        return state_dict

    @fluent
    def load_state_dict(self, state_dict, resume=True, **kwargs):
        """Resume this trial from the given state. Expects that this trial was constructed in the same way. Optionally,
        just load the model state when resume=False.

        Args:
            state_dict (dict): The state dict to reload
            resume (bool): If True, resume from the given state. Else, just load in the model weights.
            kwargs: See: `torch.nn.Module.load_state_dict <https://pytorch.org/docs/stable/nn.html?highlight=#torch.nn.Module.load_state_dict>`_

        Returns:
            :class:`.Trial`: self
        """
        if resume and torchbearer.MODEL in state_dict:  # torchbearer dict
            if torchbearer.VERSION in state_dict and state_dict[torchbearer.VERSION] != torchbearer.__version__.replace('.dev', ''):
                warnings.warn('This state dict was saved with a different torchbearer version, loading available keys. Consider setting resume=False')

            if torchbearer.MODEL in state_dict:
                self.state[torchbearer.MODEL].load_state_dict(state_dict[torchbearer.MODEL], **kwargs)

            if torchbearer.OPTIMIZER in state_dict:
                self.state[torchbearer.OPTIMIZER].load_state_dict(state_dict[torchbearer.OPTIMIZER])

            if torchbearer.HISTORY in state_dict:
                self.state[torchbearer.HISTORY] = state_dict[torchbearer.HISTORY]

            if torchbearer.CALLBACK_LIST in state_dict:
                self.state[torchbearer.CALLBACK_LIST].load_state_dict(state_dict[torchbearer.CALLBACK_LIST])
        elif torchbearer.MODEL in state_dict:
            self.state[torchbearer.MODEL].load_state_dict(state_dict[torchbearer.MODEL], **kwargs)
        else:  # something else
            warnings.warn('Not a torchbearer state dict, passing to model')
            self.state[torchbearer.MODEL].load_state_dict(state_dict, **kwargs)



from unittest import TestCase

from mock import patch, MagicMock

import torchbearer as tb
from torchbearer.callbacks import LiveLossPlot


class TestLiveLossPlot(TestCase):
    @patch('livelossplot.PlotLosses')
    def test_on_start(self, llp_mock):
        llp = LiveLossPlot(True, 1, True, False)
        llp.on_start({})
        self.assertTrue(llp_mock.call_count == 2)

    def test_on_batch(self):
        llp = LiveLossPlot(True, 1, False, False)
        llp.batch_plt = MagicMock()
        llp.plt = MagicMock()
        state = {tb.BATCH: 1, tb.METRICS: {'test': 1}}
        llp.on_step_training(state)
        llp.on_step_training(state)

        self.assertTrue(llp.batch_plt.update.call_count == 2)
        self.assertTrue(llp.plt.update.call_count == 0)

    def test_on_batch_steps(self):
        llp = LiveLossPlot(True, 2, False, False)
        llp.batch_plt = MagicMock()
        llp.plt = MagicMock()
        state = {tb.BATCH: 1, tb.METRICS: {'test': 1}}
        llp.on_step_training(state)
        state = {tb.BATCH: 2, tb.METRICS: {'test': 1}}
        llp.on_step_training(state)
        state = {tb.BATCH: 3, tb.METRICS: {'test': 1}}
        llp.on_step_training(state)
        state = {tb.BATCH: 4, tb.METRICS: {'test': 1}}
        llp.on_step_training(state)

        self.assertTrue(llp.batch_plt.draw.call_count == 2)
        self.assertTrue(llp.plt.draw.call_count == 0)

    def test_not_on_batch(self):
        llp = LiveLossPlot(False, 10, False, False)
        llp.batch_plt = MagicMock()
        llp.plt = MagicMock()
        state = {tb.BATCH: 1, tb.METRICS: {'test': 1}}
        llp.on_step_training(state)
        llp.on_step_training(state)

        self.assertTrue(llp.batch_plt.update.call_count == 0)

    def test_on_epoch(self):
        llp = LiveLossPlot(False, 10, True, False)
        llp.batch_plt = MagicMock()
        llp.plt = MagicMock()
        state = {tb.BATCH: 1, tb.METRICS: {'test': 1}}
        llp.on_end_epoch(state)
        llp.on_end_epoch(state)

        self.assertTrue(llp.batch_plt.update.call_count == 0)
        self.assertTrue(llp.plt.update.call_count == 2)

    def test_draw_once(self):
        llp = LiveLossPlot(True, 1, True, True)
        llp.batch_plt = MagicMock()
        llp.plt = MagicMock()
        state = {tb.BATCH: 1, tb.METRICS: {'test': 1}}
        llp.on_end_epoch(state)
        llp.on_end_epoch(state)
        llp.on_end(state)

        self.assertTrue(llp.plt.draw.call_count == 1)
        self.assertTrue(llp.batch_plt.draw.call_count == 1)



"""
The base metric classes exist to enable complex data flow requirements between metrics. All metrics are either instances
of :class:`.Metric` or :class:`MetricFactory`. These can then be collected in a :class:`MetricList` or a
:class:`MetricTree`. The :class:`MetricList` simply aggregates calls from a list of metrics, whereas the
:class:`MetricTree` will pass data from its root metric to each child and collect the outputs. This enables complex
running metrics and statistics, without needing to compute the underlying values more than once. Typically,
constructions of this kind should be handled using the :mod:`decorator API <.metrics.decorators>`.

..  autoclass:: torchbearer.bases.Metric
        :members:
        :undoc-members:
"""
import inspect
from torchbearer import Metric

__defaults__ = {}


def add_default(key, metric, *args, **kwargs):
    __defaults__[key] = (metric, args, kwargs)


def get_default(key):
    metric, args, kwargs = __defaults__[key]
    if inspect.isclass(metric):
        metric = metric(*args, **kwargs)
    return metric


class MetricTree(Metric):
    """A tree structure which has a node :class:`.Metric` and some children. Upon execution, the node is called with the
    input and its output is passed to each of the children. A dict is updated with the results.

    .. note::

       If the node output is already a dict (i.e. the node is a standalone metric), this is unwrapped before passing the
       **first** value to the children.

    Args:
        metric (:class:`.Metric`): The metric to act as the root node of the tree / subtree
    """
    def __init__(self, metric):
        super(MetricTree, self).__init__(metric.name)
        self.root = metric
        self.children = []

    def __str__(self):
        return str(self.root)

    def add_child(self, child):
        """Add a child to this node of the tree

        Args:
            child (:class:`.Metric`): The child to add
        """
        self.children.append(child)

    def _for_tree(self, function, *args):
        result = {}
        node_value = function(self.root, *args)
        if isinstance(node_value, dict):
            node_value = next(iter(node_value.values()))

        for subtree in self.children:
            out = function(subtree, node_value)
            if out is not None:
                result.update(out)

        return result

    def process(self, *args):
        """Process this node and then pass the output to each child.

        Returns:
            A dict containing all results from the children
        """
        return self._for_tree(lambda metric, *in_args: metric.process(*in_args), *args)

    def process_final(self, *args):
        """Process this node and then pass the output to each child.

        Returns:
            A dict containing all results from the children
        """
        return self._for_tree(lambda metric, *in_args: metric.process_final(*in_args), *args)

    def eval(self, data_key=None):
        self.root.eval(data_key=data_key)

        for subtree in self.children:
            subtree.eval(data_key=data_key)

    def train(self):
        self.root.train()

        for subtree in self.children:
            subtree.train()

    def reset(self, state):
        self.root.reset(state)

        for subtree in self.children:
            subtree.reset(state)


class MetricList(Metric):
    """The :class:`MetricList` class is a wrapper for a list of metrics which acts as a single metric and produces a
    dictionary of outputs.

    Args:
        metric_list (list): The list of metrics to be wrapped. If the list contains a :class:`MetricList`, this will be
            unwrapped. Any strings in the list will be retrieved from metrics.DEFAULT_METRICS.
    """

    def __init__(self, metric_list):
        super(MetricList, self).__init__('metric_list')

        self.metric_list = []

        for metric in metric_list:

            if isinstance(metric, str):
                metric = get_default(metric)

            if isinstance(metric, MetricList):
                self.metric_list = self.metric_list + metric.metric_list
            else:
                self.metric_list.append(metric)

    def _for_list(self, function):
        result = {}
        for metric in self.metric_list:
            out = function(metric)
            if out is not None:
                result.update(out)
        return result

    def process(self, *args):
        """Process each metric an wrap in a dictionary which maps metric names to values.

        Returns:
            dict[str,any]: A dictionary which maps metric names to values.
        """
        return self._for_list(lambda metric: metric.process(*args))

    def process_final(self, *args):
        """Process each metric an wrap in a dictionary which maps metric names to values.

        Returns:
            dict[str,any]: A dictionary which maps metric names to values.

        """
        return self._for_list(lambda metric: metric.process_final(*args))

    def train(self):
        """Put each metric in train mode.
        """
        self._for_list(lambda metric: metric.train())

    def eval(self, data_key=None):
        """Put each metric in eval mode
        """
        self._for_list(lambda metric: metric.eval(data_key=data_key))

    def reset(self, state):
        """Reset each metric with the given state.

        Args:
            state: The current state dict of the :class:`.Trial`.
        """
        self._for_list(lambda metric: metric.reset(state))

    def __str__(self):
        return str([str(m) for m in self.metric_list])


class AdvancedMetric(Metric):
    """The :class:`AdvancedMetric` class is a metric which provides different process methods for training and
    validation. This enables running metrics which do not output intermediate steps during validation.

    Args:
        name (str): The name of the metric.
    """

    def __init__(self, name):
        super(AdvancedMetric, self).__init__(name)
        self._process = self.process_train
        self._process_final = self.process_final_train

    def process_train(self, *args):
        """Process the given state and return the metric value for a training iteration.

        Returns:
            The metric value for a training iteration.
        """
        pass

    def process_validate(self, *args):
        """Process the given state and return the metric value for a validation iteration.

        Returns:
            The metric value for a validation iteration.
        """
        pass

    def process_final_train(self, *args):
        """Process the given state and return the final metric value for a training iteration.

        Returns:
            The final metric value for a training iteration.
        """
        pass

    def process_final_validate(self, *args):
        """Process the given state and return the final metric value for a validation iteration.

        Returns:
            The final metric value for a validation iteration.
        """
        pass

    def process(self, *args):
        """Depending on the current mode, return the result of either 'process_train' or 'process_validate'.

        Returns:
            The metric value.
        """
        return self._process(*args)

    def process_final(self, *args):
        """Depending on the current mode, return the result of either 'process_final_train' or 'process_final_validate'.

        Returns:
            The final metric value.
        """
        return self._process_final(*args)

    def eval(self, data_key=None):
        """Put the metric in eval mode.

        Args:
            data_key (:class:`torchbearer.StateKey`): The torchbearer data_key, if used
        """
        self._process = self.process_validate
        self._process_final = self.process_final_validate

    def train(self):
        """Put the metric in train mode.
        """
        self._process = self.process_train
        self._process_final = self.process_final_train



import torchbearer

from torchbearer.callbacks import Callback

import torch


class TorchScheduler(Callback):
    def __init__(self, scheduler_builder, monitor=None, step_on_batch=False):
        self._scheduler_builder = scheduler_builder
        self._monitor = monitor
        self._scheduler = None
        self._step_on_batch = step_on_batch

    def on_start(self, state):
        self._scheduler = self._scheduler_builder(state[torchbearer.OPTIMIZER])

    def on_sample(self, state):
        if self._step_on_batch and self._monitor is None:
            self._scheduler.step()

    def on_step_training(self, state):
        if self._step_on_batch and self._monitor is not None:
            self._scheduler.step(state[torchbearer.METRICS][self._monitor])

    def on_start_training(self, state):
        if not self._step_on_batch and self._monitor is None:
            self._scheduler.step(epoch=state[torchbearer.EPOCH])

    def on_end_epoch(self, state):
        if not self._step_on_batch and self._monitor is not None:
            self._scheduler.step(state[torchbearer.METRICS][self._monitor], epoch=state[torchbearer.EPOCH])


class LambdaLR(TorchScheduler):
    """
    Args:
        step_on_batch (bool): If True, step will be called on each training iteration rather than on each epoch

    See:
        `PyTorch LambdaLR <http://pytorch.org/docs/master/optim.html#torch.optim.lr_scheduler.LambdaLR>`_
    """
    def __init__(self, lr_lambda, last_epoch=-1, step_on_batch=False):
        super(LambdaLR, self).__init__(lambda opt:
                                       torch.optim.lr_scheduler.LambdaLR(opt, lr_lambda, last_epoch=last_epoch),
                                       step_on_batch=step_on_batch)


class StepLR(TorchScheduler):
    """
    Args:
        step_on_batch (bool): If True, step will be called on each training iteration rather than on each epoch

    See:
        `PyTorch StepLR <http://pytorch.org/docs/master/optim.html#torch.optim.lr_scheduler.StepLR>`_
    """
    def __init__(self, step_size, gamma=0.1, last_epoch=-1, step_on_batch=False):
        super(StepLR, self).__init__(lambda opt:
                                     torch.optim.lr_scheduler.StepLR(opt, step_size, gamma=gamma,
                                                                     last_epoch=last_epoch),
                                     step_on_batch=step_on_batch)


class MultiStepLR(TorchScheduler):
    """
    Args:
        step_on_batch (bool): If True, step will be called on each training iteration rather than on each epoch

    See:
        `PyTorch MultiStepLR <http://pytorch.org/docs/master/optim.html#torch.optim.lr_scheduler.MultiStepLR>`_
    """
    def __init__(self, milestones, gamma=0.1, last_epoch=-1, step_on_batch=False):
        super(MultiStepLR, self).__init__(lambda opt:
                                          torch.optim.lr_scheduler.MultiStepLR(opt, milestones, gamma=gamma,
                                                                               last_epoch=last_epoch),
                                          step_on_batch=step_on_batch)


class ExponentialLR(TorchScheduler):
    """
    Args:
        step_on_batch (bool): If True, step will be called on each training iteration rather than on each epoch

    See:
        `PyTorch ExponentialLR <http://pytorch.org/docs/master/optim.html#torch.optim.lr_scheduler.ExponentialLR>`_
    """
    def __init__(self, gamma, last_epoch=-1, step_on_batch=False):
        super(ExponentialLR, self).__init__(lambda opt:
                                            torch.optim.lr_scheduler.ExponentialLR(opt, gamma, last_epoch=last_epoch),
                                            step_on_batch=step_on_batch)


class CosineAnnealingLR(TorchScheduler):
    """
    Args:
        step_on_batch (bool): If True, step will be called on each training iteration rather than on each epoch

    See:
        `PyTorch CosineAnnealingLR <http://pytorch.org/docs/master/optim.html#torch.optim.lr_scheduler.CosineAnnealingLR>`_
    """
    def __init__(self, T_max, eta_min=0, last_epoch=-1, step_on_batch=False):
        super(CosineAnnealingLR, self).__init__(lambda opt:
                                                torch.optim.lr_scheduler.CosineAnnealingLR(opt, T_max, eta_min=eta_min,
                                                                                           last_epoch=last_epoch),
                                                step_on_batch=step_on_batch)


class ReduceLROnPlateau(TorchScheduler):
    """
    Args:
        monitor (str): The name of the quantity in metrics to monitor. (Default value = 'val_loss')
        step_on_batch (bool): If True, step will be called on each training iteration rather than on each epoch

    See:
        `PyTorch ReduceLROnPlateau <http://pytorch.org/docs/master/optim.html#torch.optim.lr_scheduler.ReduceLROnPlateau>`_
    """
    def __init__(self,  monitor='val_loss', mode='min', factor=0.1, patience=10, verbose=False, threshold=1e-4,
                 threshold_mode='rel', cooldown=0, min_lr=0, eps=1e-8, step_on_batch=False):
        super(ReduceLROnPlateau, self).__init__(lambda opt:
                                                torch.optim.lr_scheduler.ReduceLROnPlateau(
                                                    opt, mode=mode, factor=factor, patience=patience, verbose=verbose,
                                                    threshold=threshold, threshold_mode=threshold_mode,
                                                    cooldown=cooldown, min_lr=min_lr, eps=eps), monitor=monitor,
                                                step_on_batch=step_on_batch)



import torchbearer
from torchbearer.callbacks import Callback
import torch


class AggregatePredictions(Callback):
    def __init__(self):
        super(AggregatePredictions, self).__init__()
        self.predictions_list = []

    def on_step_validation(self, state):
        super(AggregatePredictions, self).on_step_validation(state)
        self.predictions_list.append(state[torchbearer.Y_PRED])

    def on_end_validation(self, state):
        state[torchbearer.FINAL_PREDICTIONS] = torch.cat(self.predictions_list, 0)

    def on_end_epoch(self, state):
        super(AggregatePredictions, self).on_end_epoch(state)
        self.predictions_list = []



import sys
if sys.version_info[0] < 3:
    def set_doc(inner, doc):
        return None  # Not simple to do in Python 2.7 so we can leave it for now, just build docs with Python 3+
else:
    def set_doc(inner, doc):
        inner.__doc__ = doc


def cite(bibtex):
    """A decorator which adds a reference to the **Google style** docstring of the given object. The ``Args:`` or
    ``Returns:`` line is then prepended with the given bibtex string at runtime. Otherwise, the last line is used.

    Args:
        bibtex (str): The bibtex string to insert

    Returns:
        The decorator
    """
    def decorator(inner):
        doc = inner.__doc__.split('\n')
        i = 0
        s = 0
        for line in doc:
            sline = line.strip()
            if sline == 'Args:' or sline == 'Returns:':
                for char in line:
                    if char == ' ':
                        s += 1
                break
            i += 1

        spaces = ' ' * (s + 4)
        to_insert = ' ' * s + '::\n\n' + spaces
        to_insert += bibtex.strip().replace('\n', '\n' + spaces).rstrip()

        doc.insert(i, '')
        doc.insert(i, to_insert)
        set_doc(inner, '\n'.join(doc))
        return inner
    return decorator



import unittest
from mock import patch, Mock

import torch

import torchbearer as tb
import torchbearer.variational.visualisation as vis


class SimpleLatentWalker(vis.LatentWalker):
    def vis(self, state):
        return None


class TestLatentWalker(unittest.TestCase):
    @patch('torchbearer.callbacks.once_per_epoch')
    def test_on_train_on_val(self, mock_ope):
        lw = SimpleLatentWalker(False, 8)
        lw.dev = 'cpu'
        lw.on_train()
        lw.on_val()

        self.assertTrue(mock_ope.call_count == 2)
        self.assertTrue(mock_ope.call_args[0][0] == lw._vis)

    def test_for_space(self):
        space_id = 3
        lw = SimpleLatentWalker(False, 8)
        lw.dev = 'cpu'
        lw.for_space(space_id)

        self.assertTrue(lw.variable_space == space_id)

    def test_data_key(self):
        data_key = 3
        lw = SimpleLatentWalker(False, 8)
        lw.dev = 'cpu'
        lw.for_data(data_key)

        self.assertTrue(lw.data_key == data_key)

    def test_to_key_to_file(self):
        out_key = 'test_key'
        out_file = 'test_file'
        lw = SimpleLatentWalker(False, 8)
        lw.dev = 'cpu'
        lw.to_key(out_key)
        lw.to_file(out_file)

        self.assertTrue(lw.file == out_file)
        self.assertTrue(lw.store_key == out_key)

    def test_vis(self):
        state = {tb.MODEL: Mock(), tb.X: Mock(), tb.DEVICE: 'cpu'}

        lw = SimpleLatentWalker(False, 8)
        lw.dev = 'cpu'
        lw.vis = Mock()
        lw._vis(state)

        self.assertTrue(lw.vis.call_count == 1)
        self.assertTrue(lw.vis.call_args[0][0] == state)

    def test_vis_to_file(self):
        state = {tb.MODEL: Mock(), tb.X: Mock(), tb.DEVICE: 'cpu'}

        lw = SimpleLatentWalker(False, 8)
        lw.to_file('test_file')
        lw.dev = 'cpu'
        lw.vis = Mock()
        lw.vis.return_value = "Test"
        lw._save_walk = Mock()
        lw._vis(state)

        self.assertTrue(lw._save_walk.call_count == 1)
        self.assertTrue(lw._save_walk.call_args[0][0] == "Test")

    def test_vis_to_key(self):
        state = {tb.MODEL: Mock(), tb.X: Mock(), tb.DEVICE: 'cpu'}

        lw = SimpleLatentWalker(False, 8)
        lw.to_key('test_key')
        lw.dev = 'cpu'
        lw.vis = Mock()
        lw.vis.return_value = "Test"
        lw._vis(state)

        self.assertTrue(state['test_key'] == "Test")

    def test_data_key_str(self):
        data_key = 'test_key'
        state = {tb.MODEL: Mock(), tb.X: Mock(), tb.DEVICE: 'cpu', data_key: 3}

        lw = SimpleLatentWalker(False, 8)
        lw.for_data(data_key)
        lw.dev = 'cpu'
        lw.vis = Mock()
        lw.vis.return_value = "Test"
        lw._vis(state)

        self.assertTrue(lw.data == state['test_key'])

    @patch('torchbearer.variational.visualisation.save_image')
    def test_save_walk(self, mock_save_img):
        state = {tb.MODEL: Mock(), tb.X: Mock(), tb.DEVICE: 'cpu'}

        lw = SimpleLatentWalker(False, 8)
        lw.to_file('test_file')
        lw.dev = 'cpu'
        lw.vis = Mock()
        lw.vis.return_value = "Test"
        lw._vis(state)

        self.assertTrue(mock_save_img.call_args[0][0] == "Test")
        self.assertTrue(mock_save_img.call_args[0][1] == "test_file")

    def test_empty_method(self):
        walker = vis.LatentWalker(False, 0)
        self.assertRaises(NotImplementedError, lambda: walker.vis({}))


class TestReconstructionViewer(unittest.TestCase):
    def test_recon(self):
        import torch
        state = {tb.X: torch.rand(4), tb.Y_PRED: torch.rand(4)}
        row_size = 2

        rv = vis.ReconstructionViewer(row_size=row_size)
        rv.data = state[tb.X]
        rv.dev = 'cpu'
        out = rv.vis(state)

        self.assertTrue((out == torch.cat([state[tb.X][:row_size], state[tb.Y_PRED][:row_size]])).all())


class TestLinspaceWalker(unittest.TestCase):
    def __init__(self, methodName):
        super(TestLinspaceWalker, self).__init__(methodName)
        import torch
        self.state = {tb.X: torch.rand(12,1,2,2), tb.Y_PRED: torch.rand(2)}

        self.codes = torch.rand(12,2)
        self.outputs = torch.rand(12,1,2,2)

        self.mock_model = Mock()
        self.mock_model.latent_dims = [2,]
        self.mock_model.encode.return_value = [self.codes]
        self.mock_model.decode.return_value = self.outputs

    def test_single_dim(self):
        lw = vis.LinSpaceWalker(dims_to_walk=[1], lin_steps=3)
        lw.data = self.state[tb.X]
        lw.model = self.mock_model
        lw.dev = 'cpu'
        lw.vis(self.state)

        correct_lin_sample = self.codes[0].repeat(3, 1)
        correct_lin_sample[0,1], correct_lin_sample[1,1], correct_lin_sample[2,1] = -1.0, 0.0, 1.0
        self.assertTrue((self.mock_model.decode.call_args[0][0][0] - correct_lin_sample[:3] < 1e-5).all())

    def test_single_dim_alt(self):
        lw = vis.LinSpaceWalker(dims_to_walk=[0], lin_steps=3)
        lw.data = self.state[tb.X]
        lw.model = self.mock_model
        lw.dev = 'cpu'
        lw.vis(self.state)

        correct_lin_sample = self.codes[0].repeat(3, 1)
        correct_lin_sample[0,0], correct_lin_sample[1,0], correct_lin_sample[2,0] = -1.0, 0.0, 1.0
        self.assertTrue((self.mock_model.decode.call_args[0][0][0] - correct_lin_sample[:3] < 1e-5).all())

    def test_single_dim_zero_init(self):
        lw = vis.LinSpaceWalker(dims_to_walk=[1], lin_steps=3, zero_init=True)
        lw.data = self.state[tb.X]
        lw.model = self.mock_model
        lw.dev = 'cpu'
        lw.vis(self.state)

        correct_lin_sample = torch.zeros(self.codes[0].shape).repeat(3, 1)
        correct_lin_sample[0,1], correct_lin_sample[1,1], correct_lin_sample[2,1] = -1.0, 0.0, 1.0
        self.assertTrue((self.mock_model.decode.call_args[0][0][0] - correct_lin_sample < 1e-5).all())

    def test_multi_dim(self):
        lw = vis.LinSpaceWalker(dims_to_walk=[0, 1], lin_steps=3)
        lw.data = self.state[tb.X]
        lw.model = self.mock_model
        lw.dev = 'cpu'
        lw.vis(self.state)

        correct_lin_sample = torch.cat([self.codes[0].repeat(3, 1), self.codes[1].repeat(3, 1)])
        correct_lin_sample[0,0], correct_lin_sample[1,0], correct_lin_sample[2,0] = -1.0, 0.0, 1.0
        correct_lin_sample[3,1], correct_lin_sample[4,1], correct_lin_sample[5,1] = -1.0, 0.0, 1.0

        self.assertTrue((self.mock_model.decode.call_args[0][0][0] - correct_lin_sample < 1e-5).all())

    def test_multi_dim_same_image(self):
        lw = vis.LinSpaceWalker(dims_to_walk=[0, 1], lin_steps=3, same_image=True)
        lw.data = self.state[tb.X]
        lw.model = self.mock_model
        lw.dev = 'cpu'
        lw.vis(self.state)

        correct_lin_sample = self.codes[0].repeat(6, 1)
        correct_lin_sample[0,0], correct_lin_sample[1,0], correct_lin_sample[2,0] = -1.0, 0.0, 1.0
        correct_lin_sample[3,1], correct_lin_sample[4,1], correct_lin_sample[5,1] = -1.0, 0.0, 1.0

        self.assertTrue((self.mock_model.decode.call_args[0][0][0] - correct_lin_sample < 1e-5).all())

    def test_multi_dim_zero_init(self):
        lw = vis.LinSpaceWalker(dims_to_walk=[0, 1], lin_steps=3, zero_init=True)
        lw.data = self.state[tb.X]
        lw.model = self.mock_model
        lw.dev = 'cpu'
        lw.vis(self.state)

        correct_lin_sample = torch.cat([torch.zeros(self.codes[0].shape).repeat(3, 1), torch.zeros(self.codes[1].shape).repeat(3, 1)])
        correct_lin_sample[0,0], correct_lin_sample[1,0], correct_lin_sample[2,0] = -1.0, 0.0, 1.0
        correct_lin_sample[3,1], correct_lin_sample[4,1], correct_lin_sample[5,1] = -1.0, 0.0, 1.0

        self.assertTrue((self.mock_model.decode.call_args[0][0][0] - correct_lin_sample < 1e-5).all())

    def test_limits(self):
        lw = vis.LinSpaceWalker(dims_to_walk=[0, 1], lin_steps=3, lin_start=-2, same_image=True)
        lw.data = self.state[tb.X]
        lw.model = self.mock_model
        lw.dev = 'cpu'
        lw.vis(self.state)

        correct_lin_sample = self.codes[0].repeat(6, 1)
        correct_lin_sample[0,0], correct_lin_sample[1,0], correct_lin_sample[2,0] = -2.0, -0.5, 1.0
        correct_lin_sample[3,1], correct_lin_sample[4,1], correct_lin_sample[5,1] = -2.0, -0.5, 1.0

        self.assertTrue((self.mock_model.decode.call_args[0][0][0] - correct_lin_sample < 1e-5).all())


class TestRandomWalker(unittest.TestCase):
    def __init__(self, methodName):
        super(TestRandomWalker, self).__init__(methodName)
        import torch
        self.state = {tb.X: torch.rand(12,1,2,2), tb.Y_PRED: torch.rand(2)}

        self.codes = torch.rand(12,2)
        self.outputs = torch.rand(32,1,2,2)

        self.mock_model = Mock()
        self.mock_model.latent_dims = [2,]
        self.mock_model.encode.return_value = [self.codes]
        self.mock_model.decode.return_value = self.outputs

    def test_code_shape(self):
        rw = vis.RandomWalker()
        rw.data = self.state[tb.X]
        rw.model = self.mock_model
        rw.dev = 'cpu'
        rw.vis(self.state)

        self.assertTrue(list(self.mock_model.decode.call_args[0][0][0].shape) == [32, 2])

    def test_code_shape_alt(self):
        self.mock_model.decode.return_value = torch.rand(10, 1, 2, 2)

        rw = vis.RandomWalker(num_images=10)
        rw.data = self.state[tb.X]
        rw.model = self.mock_model
        rw.dev = 'cpu'
        rw.vis(self.state)

        self.assertTrue(list(self.mock_model.decode.call_args[0][0][0].shape) == [10, 2])

    @patch('torch.randn')
    @patch('torch.rand')
    def test_uniform(self, mock_rand, mock_randn):
        mock_rand.return_value = torch.ones(2,2)
        mock_randn.return_value = torch.zeros(2,2)

        rw = vis.RandomWalker(num_images=10)
        rw.data = self.state[tb.X]
        rw.model = self.mock_model
        rw.dev = 'cpu'
        rw.vis(self.state)

        self.assertTrue(mock_rand.call_count == 0)
        self.assertTrue(mock_randn.call_count == 1)

        rw = vis.RandomWalker(num_images=10, uniform=True)
        rw.data = self.state[tb.X]
        rw.dev = 'cpu'
        rw.model = self.mock_model
        rw.vis(self.state)
        self.assertTrue(mock_rand.call_count == 1)
        self.assertTrue(mock_randn.call_count == 1)


class TestCodePathWalker(unittest.TestCase):
    def __init__(self, methodName):
        super(TestCodePathWalker, self).__init__(methodName)
        import torch
        self.state = {tb.X: torch.rand(12,1,2,2), tb.Y_PRED: torch.rand(2)}

        self.codes = torch.rand(12,2)
        self.outputs = torch.rand(2*5,1,2,2)

        self.mock_model = Mock()
        self.mock_model.latent_dims = [2,]
        self.mock_model.encode.return_value = [self.codes]
        self.mock_model.decode.return_value = self.outputs

    def test_vis(self):
        code1 = torch.eye(2, 2)
        code2 = (torch.eye(2, 2) == 0).to(torch.float)

        pw = vis.CodePathWalker(5, code1, code2)
        pw.data = self.state[tb.X]
        pw.model = self.mock_model
        pw.dev = 'cpu'
        pw.vis(self.state)

        correct_code = torch.zeros(10, 2)
        line = torch.Tensor([0.0, 0.25, 0.5, 0.75, 1.0])
        line_inv = torch.Tensor([1.0, 0.75, 0.5, 0.25, 0.0])
        correct_code[:5, 0], correct_code[5:, 0] = line_inv, line
        correct_code[:5, 1], correct_code[5:, 1] = line, line_inv

        self.assertTrue((self.mock_model.decode.call_args[0][0] == correct_code).all())


class TestImagePathWalker(unittest.TestCase):
    def __init__(self, methodName):
        super(TestImagePathWalker, self).__init__(methodName)
        import torch
        self.state = {tb.X: torch.rand(12,1,2,2), tb.Y_PRED: torch.rand(2)}

        self.codes = torch.rand(12,2)
        self.outputs = torch.rand(2*5,1,2,2)

        self.mock_model = Mock()
        self.mock_model.latent_dims = [2,]
        self.mock_model.encode.return_value = [self.codes]
        self.mock_model.decode.return_value = self.outputs

    @patch('torchbearer.variational.CodePathWalker.vis')
    def test_vis(self, _):
        im1 = torch.rand(2, 5, 5)
        im2 = torch.rand(2, 5, 5)
        ipw = vis.ImagePathWalker(4, im1, im2)
        ipw.data = self.state[tb.X]
        ipw.model = self.mock_model
        ipw.dev = 'cpu'
        ipw.vis(self.state)

        self.assertTrue((self.mock_model.encode.call_args_list[0][0][0] == im1).all())
        self.assertTrue((self.mock_model.encode.call_args_list[1][0][0] == im2).all())



# Sets global variable notebook when line magic is called with "%torchbearer notebook"global notebook
notebook = {'nb': False}


def set_notebook(is_notebook):
    notebook['nb'] = is_notebook


def torchbearer(line):
    if line == 'notebook':
        set_notebook(True)
    elif line == 'normal':
        set_notebook(False)


try:
    import IPython.core.magic
    torchbearer = IPython.core.magic.register_line_magic(torchbearer)
    set_notebook(True)
except NameError:
    pass


def is_notebook():
    return notebook['nb']



from unittest import TestCase

from mock import patch, MagicMock

import torchbearer
from torchbearer.callbacks import Tqdm, ConsolePrinter


class TestFormatMetrics(TestCase):
    def test_precision(self):
        metrics = {'test': 1.2345}
        res = torchbearer.callbacks.printer._format_metrics(metrics, lambda x: round(x, 3))
        self.assertEqual('test=1.234', res)

    def test_string(self):
        metrics = {'test': '1.2345'}
        res = torchbearer.callbacks.printer._format_metrics(metrics, lambda x: round(x, 3))
        self.assertEqual(res, 'test=1.2345')

    def test_not_string(self):
        metrics = {'test': {'hello': 2}}
        res = torchbearer.callbacks.printer._format_metrics(metrics, lambda x: round(x, 3))
        self.assertEqual(res, 'test={\'hello\': 2}')


class TestConsolePrinter(TestCase):
    @patch('torchbearer.callbacks.printer.print')
    def test_console_printer(self, mock_print):
        state = {torchbearer.BATCH: 5, torchbearer.EPOCH: 1, torchbearer.MAX_EPOCHS: 10, torchbearer.TRAIN_STEPS: 100, torchbearer.VALIDATION_STEPS: 101, torchbearer.METRICS: {'test': 0.99456}}
        printer = ConsolePrinter(validation_label_letter='e')
        state[torchbearer.STEPS] = state[torchbearer.TRAIN_STEPS]

        printer.on_step_training(state)
        mock_print.assert_called_once_with('\r1/10(t): 5/100 test=0.9946', end='')
        mock_print.reset_mock()

        printer.on_end_training(state)
        mock_print.assert_called_once_with('\r1/10(t): test=0.9946')
        mock_print.reset_mock()

        state[torchbearer.STEPS] = state[torchbearer.VALIDATION_STEPS]
        printer.on_step_validation(state)
        mock_print.assert_called_once_with('\r1/10(e): 5/101 test=0.9946', end='')
        mock_print.reset_mock()

        printer.on_end_validation(state)
        mock_print.assert_called_once_with('\r1/10(e): test=0.9946')
        mock_print.reset_mock()


class TestTqdm(TestCase):
    def test_tqdm(self):
        state = {torchbearer.EPOCH: 1, torchbearer.MAX_EPOCHS: 10, torchbearer.TRAIN_STEPS: 100, torchbearer.VALIDATION_STEPS: 101, torchbearer.METRICS: {'test': 0.99456}}
        tqdm = Tqdm(validation_label_letter='e')
        tqdm.tqdm_module = MagicMock()
        mock_tqdm = tqdm.tqdm_module
        state[torchbearer.STEPS] = state[torchbearer.TRAIN_STEPS]

        tqdm.on_start_training(state)
        mock_tqdm.assert_called_once_with(total=100, desc='1/10(t)')

        tqdm.on_step_training(state)
        mock_tqdm.return_value.set_postfix_str.assert_called_once_with('test=0.9946')
        mock_tqdm.return_value.update.assert_called_once_with(1)
        mock_tqdm.return_value.set_postfix_str.reset_mock()

        tqdm.on_end_training(state)
        mock_tqdm.return_value.set_postfix_str.assert_called_once_with('test=0.9946')
        self.assertEqual(mock_tqdm.return_value.close.call_count, 1)

        mock_tqdm.reset_mock()
        mock_tqdm.return_value.set_postfix_str.reset_mock()
        mock_tqdm.return_value.update.reset_mock()
        mock_tqdm.return_value.close.reset_mock()

        state[torchbearer.STEPS] = state[torchbearer.VALIDATION_STEPS]
        tqdm.on_start_validation(state)
        mock_tqdm.assert_called_once_with(total=101, desc='1/10(e)')

        tqdm.on_step_validation(state)
        mock_tqdm.return_value.set_postfix_str.assert_called_once_with('test=0.9946')
        mock_tqdm.return_value.update.assert_called_once_with(1)
        mock_tqdm.return_value.set_postfix_str.reset_mock()

        tqdm.on_end_validation(state)
        mock_tqdm.return_value.set_postfix_str.assert_called_once_with('test=0.9946')
        self.assertEqual(mock_tqdm.return_value.close.call_count, 1)

    def test_tqdm_custom_args(self):
        state = {torchbearer.EPOCH: 1, torchbearer.MAX_EPOCHS: 10, torchbearer.TRAIN_STEPS: 100,
                 torchbearer.VALIDATION_STEPS: 101, torchbearer.METRICS: {'test': 10}}
        state[torchbearer.HISTORY] = [[[], state[torchbearer.METRICS]]]
        tqdm = Tqdm(ascii=True)
        state[torchbearer.STEPS] = state[torchbearer.TRAIN_STEPS]

        tqdm.tqdm_module = MagicMock()
        mock_tqdm = tqdm.tqdm_module

        tqdm.on_start_training(state)
        mock_tqdm.assert_called_once_with(total=100, desc='1/10(t)', ascii=True)

        tqdm = Tqdm(on_epoch=True, ascii=True)

        tqdm.tqdm_module = MagicMock()
        mock_tqdm = tqdm.tqdm_module

        tqdm.on_start(state)
        mock_tqdm.assert_called_once_with(initial=1, total=10, ascii=True)

    def test_tqdm_on_epoch(self):
        state = {torchbearer.EPOCH: 1, torchbearer.MAX_EPOCHS: 10, torchbearer.HISTORY: [0, (1, {'test': 0.99456})],
                 torchbearer.METRICS: {'test': 0.99456}}
        tqdm = Tqdm(validation_label_letter='e', on_epoch=True)
        tqdm.tqdm_module = MagicMock()
        mock_tqdm = tqdm.tqdm_module

        tqdm.on_start(state)
        mock_tqdm.assert_called_once_with(initial=2, total=10)
        mock_tqdm.return_value.set_postfix_str.assert_called_once_with('test=0.9946')
        mock_tqdm.return_value.update.assert_called_once_with(1)
        mock_tqdm.return_value.set_postfix_str.reset_mock()
        mock_tqdm.return_value.update.reset_mock()

        tqdm.on_end_epoch(state)
        mock_tqdm.return_value.set_postfix_str.assert_called_once_with('test=0.9946')
        mock_tqdm.return_value.update.assert_called_once_with(1)
        mock_tqdm.return_value.set_postfix_str.reset_mock()

        tqdm.on_end(state)
        mock_tqdm.return_value.set_postfix_str.assert_called_once_with('test=0.9946')
        self.assertEqual(mock_tqdm.return_value.close.call_count, 1)

    @patch('torchbearer.magics.is_notebook')
    def test_tqdm_module_init_notebook(self, mock_is_notebook):
        from tqdm import tqdm_notebook
        mock_is_notebook.return_value = True
        tqdm = Tqdm(validation_label_letter='e', on_epoch=True)
        self.assertTrue(tqdm.tqdm_module == tqdm_notebook)

    @patch('torchbearer.magics.is_notebook')
    def test_tqdm_module_init_notebook(self, mock_is_notebook):
        from tqdm import tqdm as base_tqdm
        mock_is_notebook.return_value = False
        tqdm = Tqdm(validation_label_letter='e', on_epoch=True)
        self.assertTrue(tqdm.tqdm_module == base_tqdm)



"""
..  autoclass:: torchbearer.bases.Callback
        :members:
        :undoc-members:
"""

from torchbearer import Callback


class CallbackList(Callback):
    """The :class:`CallbackList` class is a wrapper for a list of callbacks which acts as a single :class:`.Callback` and
    internally calls each :class:`.Callback` in the given list in turn.

    Args:
        callback_list (list): The list of callbacks to be wrapped. If the list contains a :class:`CallbackList`, this
            will be unwrapped.
    """

    CALLBACK_STATES = 'callback_states'
    CALLBACK_TYPES = 'callback_types'

    def __init__(self, callback_list):
        super(CallbackList, self).__init__()
        self.callback_list = []
        self.append(callback_list)

    def state_dict(self):
        """Get a dict containing all of the callback states.

        Returns:
            dict: A dict containing parameters and persistent buffers.
        """
        state_dict = {
            CallbackList.CALLBACK_STATES: [],
            CallbackList.CALLBACK_TYPES: []
        }

        def to_state(callback):
            state_dict[CallbackList.CALLBACK_STATES].append(callback.state_dict())
            state_dict[CallbackList.CALLBACK_TYPES].append(callback.__class__)

        self._for_list(to_state)

        return state_dict

    def load_state_dict(self, state_dict):
        """Resume this callback list from the given state. Callbacks must be given in the same order for this to work.

        Args:
            state_dict (dict): The state dict to reload

        Returns:
            CallbackList: self
        """

        t_iter = iter(state_dict[CallbackList.CALLBACK_TYPES])
        s_iter = iter(state_dict[CallbackList.CALLBACK_STATES])

        def from_state(callback):
            if callback.__class__ == next(t_iter):
                callback.load_state_dict(next(s_iter))
            else:
                import warnings
                warnings.warn('Callback classes did not match, expected: ' + str([c.__name__ for c in state_dict[CallbackList.CALLBACK_TYPES]]))

        self._for_list(from_state)

        return self

    def _for_list(self, function):
        for callback in self.callback_list:
            function(callback)

    def __str__(self):
        return str([str(c) for c in self.callback_list])

    def __iter__(self):
        return self.callback_list.__iter__()

    def __copy__(self):
        return CallbackList(self.callback_list)

    def copy(self):
        return self.__copy__()

    def append(self, callback_list):
        for callback in callback_list:
            if isinstance(callback, CallbackList):
                self.callback_list = self.callback_list + callback.callback_list
            else:
                self.callback_list.append(callback)
        
    def on_start(self, state):
        """Call on_start on each callback in turn with the given state.

        Args:
            state (dict[str,any]): The current state dict of the :class:`.Trial`.
        """
        self._for_list(lambda callback: callback.on_start(state))

    def on_start_epoch(self, state):
        """Call on_start_epoch on each callback in turn with the given state.

        Args:
            state (dict[str,any]): The current state dict of the :class:`.Trial`.
        """
        self._for_list(lambda callback: callback.on_start_epoch(state))

    def on_start_training(self, state):
        """Call on_start_training on each callback in turn with the given state.

        Args:
            state (dict[str,any]): The current state dict of the :class:`.Trial`.
        """
        self._for_list(lambda callback: callback.on_start_training(state))

    def on_sample(self, state):
        """Call on_sample on each callback in turn with the given state.

        Args:
            state (dict[str,any]): The current state dict of the :class:`.Trial`.
        """
        self._for_list(lambda callback: callback.on_sample(state))

    def on_forward(self, state):
        """Call on_forward on each callback in turn with the given state.

        Args:
            state (dict[str,any]): The current state dict of the :class:`.Trial`.
        """
        self._for_list(lambda callback: callback.on_forward(state))

    def on_criterion(self, state):
        """Call on_criterion on each callback in turn with the given state.

        Args:
            state (dict[str,any]): The current state dict of the :class:`.Trial`.
        """
        self._for_list(lambda callback: callback.on_criterion(state))

    def on_backward(self, state):
        """Call on_backward on each callback in turn with the given state.

        Args:
            state (dict[str,any]): The current state dict of the :class:`.Trial`.
        """
        self._for_list(lambda callback: callback.on_backward(state))

    def on_step_training(self, state):
        """Call on_step_training on each callback in turn with the given state.

        Args:
            state (dict[str,any]): The current state dict of the :class:`.Trial`.
        """
        self._for_list(lambda callback: callback.on_step_training(state))

    def on_end_training(self, state):
        """Call on_end_training on each callback in turn with the given state.

        Args:
            state (dict[str,any]): The current state dict of the :class:`.Trial`.
        """
        self._for_list(lambda callback: callback.on_end_training(state))

    def on_end_epoch(self, state):
        """Call on_end_epoch on each callback in turn with the given state.

        Args:
            state (dict[str,any]): The current state dict of the :class:`.Trial`.
        """
        self._for_list(lambda callback: callback.on_end_epoch(state))

    def on_checkpoint(self, state):
        """Call on_checkpoint on each callback in turn with the given state.

        Args:
            state (dict[str,any]): The current state dict of the :class:`.Trial`.
        """
        self._for_list(lambda callback: callback.on_checkpoint(state))

    def on_end(self, state):
        """Call on_end on each callback in turn with the given state.

        Args:
            state (dict[str,any]): The current state dict of the :class:`.Trial`.
        """
        self._for_list(lambda callback: callback.on_end(state))

    def on_start_validation(self, state):
        """Call on_start_validation on each callback in turn with the given state.

        Args:
            state (dict[str,any]): The current state dict of the :class:`.Trial`.
        """
        self._for_list(lambda callback: callback.on_start_validation(state))

    def on_sample_validation(self, state):
        """Call on_sample_validation on each callback in turn with the given state.

        Args:
            state (dict[str,any]): The current state dict of the :class:`.Trial`.
        """
        self._for_list(lambda callback: callback.on_sample_validation(state))

    def on_forward_validation(self, state):
        """Call on_forward_validation on each callback in turn with the given state.

        Args:
            state (dict[str,any]): The current state dict of the :class:`.Trial`.
        """
        self._for_list(lambda callback: callback.on_forward_validation(state))

    def on_criterion_validation(self, state):
        """Call on_criterion_validation on each callback in turn with the given state.

        Args:
            state (dict[str,any]): The current state dict of the :class:`.Trial`.
        """
        self._for_list(lambda callback: callback.on_criterion_validation(state))

    def on_end_validation(self, state):
        """Call on_end_validation on each callback in turn with the given state.

        Args:
            state (dict[str,any]): The current state dict of the :class:`.Trial`.
        """
        self._for_list(lambda callback: callback.on_end_validation(state))

    def on_step_validation(self, state):
        """Call on_step_validation on each callback in turn with the given state.

        Args:
            state (dict[str,any]): The current state dict of the :class:`.Trial`.
        """
        self._for_list(lambda callback: callback.on_step_validation(state))



import copy
import os

import torch
import torch.nn.functional as F
import torchvision.utils as utils

import torchbearer
from torchbearer.callbacks import Callback

__writers__ = dict()


class VisdomParams:
    """
    Class to hold visdom client arguments. Modify member variables before initialising tensorboard callbacks for custom
    arguments. See: `visdom <https://github.com/facebookresearch/visdom#visdom-arguments-python-only>`_
    """
    SERVER = 'http://localhost'
    ENDPOINT = 'events'
    PORT = 8097
    IPV6 = True
    HTTP_PROXY_HOST = None
    HTTP_PROXY_PORT = None
    ENV = 'main'
    SEND = True
    RAISE_EXCEPTIONS = None
    USE_INCOMING_SOCKET = True
    LOG_TO_FILENAME = None

    def __to_dict__(self):
        base_params = {e.lower(): VisdomParams.__dict__[e] for e in VisdomParams.__dict__ if '__' not in e}
        new_params = {e.lower(): self.__dict__[e] for e in self.__dict__ if '__' not in e}
        base_params.update(new_params)
        return base_params


def get_writer(log_dir, logger, visdom=False, visdom_params=None):
    """
    Get the writer assigned to the given log directory.
    If the writer doesn't exist it will be created, and a reference to the logger added.

    Args:
        log_dir (str): the log directory
        logger: the object requesting the writer. That object should call `close_writer` when its finished
        visdom (bool): if true VisdomWriter is returned instead of tensorboard SummaryWriter
        visdom_params (:class:`.VisdomParams`): Visdom parameter settings object, uses default if None

    Returns:
        the `SummaryWriter` or `VisdomWriter` object
    """
    import tensorboardX
    from tensorboardX import SummaryWriter
    import sys
    import errno
    kwargs = {}
    if sys.version_info[0] >= 3:
        kwargs['exist_ok'] = True

    writer_key = 'writer'
    if visdom:
        writer_key = 'writer_visdom'

    if log_dir not in __writers__ or writer_key not in __writers__[log_dir]:
        if visdom:
            w = tensorboardX.torchvis.VisdomWriter()
            from visdom import Visdom
            try:
                os.makedirs(log_dir, **kwargs)
            except OSError as exc:
                if exc.errno == errno.EEXIST and os.path.isdir(log_dir):
                    pass
                else:
                    raise exc
            if visdom_params is None:
                visdom_params = VisdomParams()
                visdom_params.LOG_TO_FILENAME = os.path.join(log_dir, 'log.log')
            w.vis = Visdom(**visdom_params.__to_dict__())
        else:
            w = SummaryWriter(log_dir=log_dir)
        __writers__[log_dir] = {writer_key: w, 'references': set()}

    __writers__[log_dir]['references'].add(logger)
    return __writers__[log_dir][writer_key]


def close_writer(log_dir, logger):
    """
    Decrement the reference count for a writer belonging to a specific log directory.
    If the reference count gets to zero, the writer will be closed and removed.

    Args:
        log_dir (str): the log directory
        logger: the object releasing the writer
    """
    if log_dir in __writers__:
        __writers__[log_dir]['references'].discard(logger)

        if len(__writers__[log_dir]['references']) is 0:
            if 'writer' in __writers__[log_dir]:
                __writers__[log_dir]['writer'].close()

            if 'writer_visdom' in __writers__[log_dir]:
                __writers__[log_dir]['writer_visdom'].close()

        del __writers__[log_dir]


class AbstractTensorBoard(Callback):
    """TensorBoard callback which writes metrics to the given log directory. Requires the TensorboardX library for python.

    Args:
        log_dir (str): The tensorboard log path for output
        comment (str): Descriptive comment to append to path
        visdom (bool): If true, log to visdom instead of tensorboard
        visdom_params (:class:`.VisdomParams`): Visdom parameter settings object, uses default if None
    """

    def __init__(self, log_dir='./logs',
                 comment='torchbearer', visdom=False, visdom_params=None):
        super(AbstractTensorBoard, self).__init__()

        self.raw_log_dir = log_dir
        self.log_dir = log_dir
        self.comment = comment
        self.writer = None
        self.visdom = visdom
        self.visdom_params = visdom_params

    def get_writer(self, log_dir=None, visdom=False, visdom_params=None):
        """
        Get a SummaryWriter for the given directory (or the default writer if the directory is not given).
        If you are getting a `SummaryWriter` for a custom directory, it is your responsibility to close
        it using `close_writer`.

        Args:
            log_dir (str): the (optional) directory
            visdom (bool): If true, return VisdomWriter, if false return tensorboard SummaryWriter
            visdom_params (:class:`.VisdomParams`): Visdom parameter settings object, uses default if None

        Returns:
            the `SummaryWriter` or `VisdomWriter`
        """
        if log_dir is None:
            self.writer = get_writer(self.log_dir, self, visdom=visdom, visdom_params=visdom_params)
            return self.writer
        else:
            return get_writer(log_dir, self, visdom=visdom, visdom_params=visdom_params)

    def close_writer(self, log_dir=None):
        """
        Decrement the reference count for a writer belonging to the given log directory
        (or the default writer if the directory is not given). If the reference count gets to zero,
        the writer will be closed and removed.

        Args:
            log_dir (str): the (optional) directory
        """
        if log_dir is None:
            close_writer(self.log_dir, self)
        else:
            close_writer(log_dir, self)

    def on_start(self, state):
        self.log_dir = os.path.join(self.log_dir, state[torchbearer.MODEL].__class__.__name__ + '_' + self.comment)
        self.writer = self.get_writer(visdom=self.visdom, visdom_params=self.visdom_params)

    def on_end(self, state):
        self.close_writer()


class TensorBoard(AbstractTensorBoard):
    """TensorBoard callback which writes metrics to the given log directory. Requires the TensorboardX library for python.

    Args:
        log_dir (str): The tensorboard log path for output
        write_graph (bool): If True, the model graph will be written using the TensorboardX library
        write_batch_metrics (bool): If True, batch metrics will be written
        batch_step_size (int): The step size to use when writing batch metrics, make this larger to reduce latency
        write_epoch_metrics (bool): If True, metrics from the end of the epoch will be written
        comment (str): Descriptive comment to append to path
        visdom (bool): If true, log to visdom instead of tensorboard
        visdom_params (:class:`.VisdomParams`): Visdom parameter settings object, uses default if None
    """

    def __init__(self, log_dir='./logs',
                 write_graph=True,
                 write_batch_metrics=False,
                 batch_step_size=10,
                 write_epoch_metrics=True,
                 comment='torchbearer',
                 visdom=False,
                 visdom_params=None):
        super(TensorBoard, self).__init__(log_dir, comment, visdom, visdom_params)

        self.write_graph = write_graph
        self.write_batch_metrics = write_batch_metrics
        self.batch_step_size = batch_step_size
        self.write_epoch_metrics = write_epoch_metrics
        self.visdom = visdom

        if self.write_graph and not visdom:
            def handle_graph(state):
                dummy = torch.rand(state[torchbearer.X].size(), requires_grad=False)
                model = copy.deepcopy(state[torchbearer.MODEL]).to('cpu')
                self.writer.add_graph(model, (dummy,))
                self._handle_graph = lambda _: None

            self._handle_graph = handle_graph
        else:
            self._handle_graph = lambda _: None

        self.batch_log_dir = None
        self.batch_writer = None

    def on_start_epoch(self, state):
        if self.write_batch_metrics:
            if self.visdom:
                self.batch_log_dir = os.path.join(self.log_dir, 'epoch/')
            else:
                self.batch_log_dir = os.path.join(self.log_dir, 'epoch-' + str(state[torchbearer.EPOCH]))
            self.batch_writer = self.get_writer(self.batch_log_dir, visdom=self.visdom)

    def on_sample(self, state):
        self._handle_graph(state)

    def on_step_training(self, state):
        if self.write_batch_metrics and state[torchbearer.BATCH] % self.batch_step_size == 0:
            for metric in state[torchbearer.METRICS]:
                if self.visdom:
                    self.batch_writer.add_scalar(metric, state[torchbearer.METRICS][metric],
                                                 state[torchbearer.EPOCH] * state[torchbearer.TRAIN_STEPS] + state[
                                                     torchbearer.BATCH], main_tag='batch')
                else:
                    self.batch_writer.add_scalar('batch/' + metric, state[torchbearer.METRICS][metric], state[torchbearer.BATCH])

    def on_step_validation(self, state):
        if self.write_batch_metrics and state[torchbearer.BATCH] % self.batch_step_size == 0:
            for metric in state[torchbearer.METRICS]:
                if self.visdom:
                    self.batch_writer.add_scalar(metric, state[torchbearer.METRICS][metric],
                                                 state[torchbearer.EPOCH] * state[torchbearer.TRAIN_STEPS] + state[
                                                     torchbearer.BATCH], main_tag='batch')
                else:
                    self.batch_writer.add_scalar('batch/' + metric, state[torchbearer.METRICS][metric], state[torchbearer.BATCH])

    def on_end_epoch(self, state):
        if self.write_batch_metrics and not self.visdom:
            self.close_writer(self.batch_log_dir)

        if self.write_epoch_metrics:
            for metric in state[torchbearer.METRICS]:
                if self.visdom:
                    self.writer.add_scalar(metric, state[torchbearer.METRICS][metric], state[torchbearer.EPOCH],
                                           main_tag='epoch')
                else:
                    self.writer.add_scalar('epoch/' + metric, state[torchbearer.METRICS][metric], state[torchbearer.EPOCH])

    def on_end(self, state):
        super(TensorBoard, self).on_end(state)
        if self.write_batch_metrics and self.visdom:
            self.close_writer(self.batch_log_dir)


class TensorBoardText(AbstractTensorBoard):
    """TensorBoard callback which writes metrics as text to the given log directory. Requires the TensorboardX library
    for python.


    Args:
        log_dir (str): The tensorboard log path for output
        write_epoch_metrics (bool): If True, metrics from the end of the epoch will be written
        log_trial_summary (bool): If True logs a string summary of the Trial
        batch_step_size (int): The step size to use when writing batch metrics, make this larger to reduce latency
        comment (str): Descriptive comment to append to path
        visdom (bool): If true, log to visdom instead of tensorboard
        visdom_params (:class:`.VisdomParams`): Visdom parameter settings object, uses default if None
    """

    def __init__(self, log_dir='./logs',
                 write_epoch_metrics=True,
                 write_batch_metrics=False,
                 log_trial_summary=True,
                 batch_step_size=100,
                 comment='torchbearer',
                 visdom=False,
                 visdom_params=None):
        super(TensorBoardText, self).__init__(log_dir, comment, visdom, visdom_params)
        self.write_epoch_metrics = write_epoch_metrics
        self.write_batch_metrics = write_batch_metrics
        self.log_trial_summary = log_trial_summary
        self.batch_step_size = batch_step_size
        self.visdom = visdom
        self.batch_log_dir = None
        self.batch_writer = None
        self.logged_summary = False

    @staticmethod
    def table_formatter(string):
        table = '<table><th>Metric</th><th>Value</th>'
        string = string.replace('{', '').replace('}', '').replace("'", "")  # TODO: Replace this with single pass regex

        def cell(string):
            return '<td>' + string + '</td>'

        def row(string):
            return '<tr>' + string + '</tr>'

        metrics = string.split(',')
        for _, metric in enumerate(metrics):
            items = metric.split(':')
            name, value = items[0], items[1]
            table = table + row(cell(name) + cell(value))

        return table + '</table>'

    def on_start(self, state):
        super(TensorBoardText, self).on_start(state)
        if self.log_trial_summary and not self.logged_summary:
            self.logged_summary = True
            self.writer.add_text('trial', str(state[torchbearer.SELF]).replace('\n', '\n \n'), 1)

    def on_start_epoch(self, state):
        if self.write_batch_metrics:
            if self.visdom:
                self.batch_log_dir = os.path.join(self.log_dir, 'epoch/')
                batch_params = self.visdom_params if self.visdom_params is not None else VisdomParams()
                batch_params.ENV = batch_params.ENV + '-batch'
                self.batch_writer = self.get_writer(self.batch_log_dir, visdom=self.visdom, visdom_params=batch_params)
            else:
                self.batch_log_dir = os.path.join(self.log_dir, 'epoch-' + str(state[torchbearer.EPOCH]))
                self.batch_writer = self.get_writer(self.batch_log_dir)

    def on_step_training(self, state):
        if self.write_batch_metrics and state[torchbearer.BATCH] % self.batch_step_size == 0:
            if self.visdom:
                self.batch_writer.add_text('batch', '<h3>Epoch {} - Batch {}</h3>'.format(state[torchbearer.EPOCH], state[torchbearer.BATCH])+self.table_formatter(str(state[torchbearer.METRICS])), 1)
            else:
                self.batch_writer.add_text('batch', self.table_formatter(str(state[torchbearer.METRICS])), state[torchbearer.BATCH])

    def on_end_epoch(self, state):
        if self.write_epoch_metrics:
            if self.visdom:
                self.writer.add_text('epoch', '<h4>Epoch {}</h4>'.format(state[torchbearer.EPOCH])+self.table_formatter(str(state[torchbearer.METRICS])), 1)
            else:
                self.writer.add_text('epoch', self.table_formatter(str(state[torchbearer.METRICS])), state[torchbearer.EPOCH])

    def on_end(self, state):
        super(TensorBoardText, self).on_end(state)
        if self.write_batch_metrics and self.visdom:
            self.close_writer(self.batch_log_dir)


class TensorBoardImages(AbstractTensorBoard):
    """The TensorBoardImages callback will write a selection of images from the validation pass to tensorboard using the
    TensorboardX library and torchvision.utils.make_grid. Images are selected from the given key and saved to the given
    path. Full name of image sub directory will be model name + _ + comment.

    Args:
        log_dir (str): The tensorboard log path for output
        comment (str): Descriptive comment to append to path
        name (str): The name of the image
        key (str): The key in state containing image data (tensor of size [c, w, h] or [b, c, w, h])
        write_each_epoch (bool): If True, write data on every epoch, else write only for the first epoch.
        num_images (int): The number of images to write
        nrow: See `torchvision.utils.make_grid <https://pytorch.org/docs/stable/torchvision/utils.html#torchvision.utils.make_grid>`_
        padding: See `torchvision.utils.make_grid <https://pytorch.org/docs/stable/torchvision/utils.html#torchvision.utils.make_grid>`_
        normalize: See `torchvision.utils.make_grid <https://pytorch.org/docs/stable/torchvision/utils.html#torchvision.utils.make_grid>`_
        norm_range: See `torchvision.utils.make_grid <https://pytorch.org/docs/stable/torchvision/utils.html#torchvision.utils.make_grid>`_
        scale_each: See `torchvision.utils.make_grid <https://pytorch.org/docs/stable/torchvision/utils.html#torchvision.utils.make_grid>`_
        pad_value: See `torchvision.utils.make_grid <https://pytorch.org/docs/stable/torchvision/utils.html#torchvision.utils.make_grid>`_
        visdom (bool): If true, log to visdom instead of tensorboard
        visdom_params (:class:`.VisdomParams`): Visdom parameter settings object, uses default if None
    """

    def __init__(self, log_dir='./logs',
                 comment='torchbearer',
                 name='Image',
                 key=torchbearer.Y_PRED,
                 write_each_epoch=True,
                 num_images=16,
                 nrow=8,
                 padding=2,
                 normalize=False,
                 norm_range=None,
                 scale_each=False,
                 pad_value=0,
                 visdom=False,
                 visdom_params=None):
        super(TensorBoardImages, self).__init__(log_dir, comment, visdom=visdom, visdom_params=visdom_params)
        self.name = name
        self.key = key
        self.write_each_epoch = write_each_epoch
        self.num_images = num_images
        self.nrow = nrow
        self.padding = padding
        self.normalize = normalize
        self.norm_range = norm_range
        self.scale_each = scale_each
        self.pad_value = pad_value

        self._data = None
        self.done = False

    def on_step_validation(self, state):
        if not self.done:
            data = state[self.key].clone()

            if len(data.size()) == 3:
                data = data.unsqueeze(1)

            if self._data is None:
                remaining = self.num_images if self.num_images < data.size(0) else data.size(0)

                self._data = data[:remaining].to('cpu')
            else:
                remaining = self.num_images - self._data.size(0)

                if remaining > data.size(0):
                    remaining = data.size(0)

                self._data = torch.cat((self._data, data[:remaining].to('cpu')), dim=0)

            if self._data.size(0) >= self.num_images:
                image = utils.make_grid(
                    self._data,
                    nrow=self.nrow,
                    padding=self.padding,
                    normalize=self.normalize,
                    range=self.norm_range,
                    scale_each=self.scale_each,
                    pad_value=self.pad_value
                )
                if self.visdom:
                    name = self.name + str(state[torchbearer.EPOCH])
                else:
                    name = self.name
                self.writer.add_image(name, image, state[torchbearer.EPOCH])
                self.done = True
                self._data = None

    def on_end_epoch(self, state):
        if self.write_each_epoch:
            self.done = False


class TensorBoardProjector(AbstractTensorBoard):
    """The TensorBoardProjector callback is used to write images from the validation pass to Tensorboard using the
    TensorboardX library. Images are written to the given directory and, if required, so are associated features.

    Args:
        log_dir (str): The tensorboard log path for output
        comment (str): Descriptive comment to append to path
        num_images (int): The number of images to write
        avg_pool_size (int): Size of the average pool to perform on the image. This is recommended to reduce the overall
            image sizes and improve latency
        avg_data_channels (bool): If True, the image data will be averaged in the channel dimension
        write_data (bool): If True, the raw data will be written as an embedding
        write_features (bool): If True, the image features will be written as an embedding
        features_key (str): The key in state to use for the embedding. Typically model output but can be used to show
            features from any layer of the model.
    """

    def __init__(self, log_dir='./logs',
                 comment='torchbearer',
                 num_images=100,
                 avg_pool_size=1,
                 avg_data_channels=True,
                 write_data=True,
                 write_features=True,
                 features_key=torchbearer.Y_PRED):
        super(TensorBoardProjector, self).__init__(log_dir, comment)
        self.num_images = num_images
        self.avg_pool_size = avg_pool_size
        self.avg_data_channels = avg_data_channels
        self.write_data = write_data
        self.write_features = write_features
        self.features_key = features_key
        self.done = False
        self._images = None
        self._labels = None
        self._features = None
        self._data = None

    def on_step_validation(self, state):
        if not self.done:
            x = state[torchbearer.X].data.clone()

            if len(x.size()) == 3:
                x = x.unsqueeze(1)

            x = F.avg_pool2d(x, self.avg_pool_size).data

            data = None

            if state[torchbearer.EPOCH] == 0 and self.write_data:
                if self.avg_data_channels:
                    data = torch.mean(x, 1)
                else:
                    data = x

                data = data.view(data.size(0), -1)

            feature = None

            if self.write_features:
                feature = state[self.features_key].data.clone()
                feature = feature.view(feature.size(0), -1)

            label = state[torchbearer.Y_TRUE].data.clone()

            if state[torchbearer.BATCH] == 0:
                remaining = self.num_images if self.num_images < label.size(0) else label.size(0)

                self._images = x[:remaining].to('cpu')
                self._labels = label[:remaining].to('cpu')

                if data is not None:
                    self._data = data[:remaining].to('cpu')

                if feature is not None:
                    self._features = feature[:remaining].to('cpu')
            else:
                remaining = self.num_images - self._labels.size(0)

                if remaining > label.size(0):
                    remaining = label.size(0)

                self._images = torch.cat((self._images, x[:remaining].to('cpu')), dim=0)
                self._labels = torch.cat((self._labels, label[:remaining].to('cpu')), dim=0)

                if data is not None:
                    self._data = torch.cat((self._data, data[:remaining].to('cpu')), dim=0)

                if feature is not None:
                    self._features = torch.cat((self._features, feature[:remaining].to('cpu')), dim=0)

            if self._labels.size(0) >= self.num_images:
                if state[torchbearer.EPOCH] == 0 and self.write_data:
                    self.writer.add_embedding(self._data, metadata=self._labels, label_img=self._images, tag='data',
                                              global_step=-1)
                if self.write_features:
                    self.writer.add_embedding(self._features, metadata=self._labels, label_img=self._images,
                                              tag='features', global_step=state[torchbearer.EPOCH])
                self.done = True

    def on_end_epoch(self, state):
        if self.write_features:
            self.done = False



"""
Base Classes
------------------------------------

..  automodule:: torchbearer.metrics.metrics
        :members:
        :undoc-members:

Decorators - The Decorator API
------------------------------------

..  automodule:: torchbearer.metrics.decorators
        :members:
        :undoc-members:

Metric Wrappers
------------------------------------

..  automodule:: torchbearer.metrics.wrappers
        :members:
        :undoc-members:

Metric Aggregators
------------------------------------

..  automodule:: torchbearer.metrics.aggregators
        :members:
        :undoc-members:

Base Metrics
------------------------------------

..  automodule:: torchbearer.metrics.default
        :members:

..  automodule:: torchbearer.metrics.primitives
        :members:

..  automodule:: torchbearer.metrics.roc_auc_score
        :members:


Timer
------------------------------------
..  automodule:: torchbearer.metrics.timer
        :members:
        :undoc-members:
"""

from torchbearer import Metric
from .metrics import *
from .wrappers import *
from .aggregators import *
from .decorators import *
from .roc_auc_score import *
from .primitives import *
from .timer import TimerMetric
from .default import DefaultAccuracy
from .lr import LR



"""
Trial
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

..  automodule:: torchbearer.trial
        :members:
        :undoc-members:

Model (Deprecated)
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

..  automodule:: torchbearer.torchbearer
        :members:
        :undoc-members:

State
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

..  automodule:: torchbearer.state
        :members:
        :undoc-members:

Utilities
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

..  automodule:: torchbearer.cv_utils
        :members:
        :undoc-members:
"""

from .version import __version__
from . import magics
from .cite import cite
from .bases import *
from .state import *

from . import metrics
from . import callbacks
from .trial import *
from . import cv_utils
from . import variational



from unittest import TestCase

import torchbearer
from torchbearer.callbacks import EarlyStopping


class TestEarlyStopping(TestCase):
    
    def test_min_should_stop(self):
        state = {
            torchbearer.EPOCH: 1,
            torchbearer.STOP_TRAINING: False,
            torchbearer.METRICS: {'test_metric': 0.001}
        }

        stopper = EarlyStopping(monitor='test_metric', mode='min')

        stopper.on_start(state)
        stopper.on_end_epoch(state)

        self.assertFalse(state[torchbearer.STOP_TRAINING])

        state[torchbearer.METRICS]['test_metric'] = 0.01
        stopper.on_end_epoch(state)

        self.assertTrue(state[torchbearer.STOP_TRAINING])

    def test_min_should_continue(self):
        state = {
            torchbearer.EPOCH: 1,
            torchbearer.STOP_TRAINING: False,
            torchbearer.METRICS: {'test_metric': 0.001}
        }

        stopper = EarlyStopping(monitor='test_metric', mode='min')

        stopper.on_start(state)
        stopper.on_end_epoch(state)

        self.assertFalse(state[torchbearer.STOP_TRAINING])

        state[torchbearer.METRICS]['test_metric'] = 0.0001

        stopper.on_end_epoch(state)

        self.assertFalse(state[torchbearer.STOP_TRAINING])

    def test_max_should_stop(self):
        state = {
            torchbearer.EPOCH: 1,
            torchbearer.STOP_TRAINING: False,
            torchbearer.METRICS: {'test_metric': 0.001}
        }

        stopper = EarlyStopping(monitor='test_metric',  mode='max')

        stopper.on_start(state)
        stopper.on_end_epoch(state)

        self.assertFalse(state[torchbearer.STOP_TRAINING])

        state[torchbearer.METRICS]['test_metric'] = 0.0001
        stopper.on_end_epoch(state)

        self.assertTrue(state[torchbearer.STOP_TRAINING])

    def test_max_should_continue(self):
        state = {
            torchbearer.EPOCH: 1,
            torchbearer.STOP_TRAINING: False,
            torchbearer.METRICS: {'test_metric': 0.001}
        }

        stopper = EarlyStopping(monitor='test_metric', mode='max')

        stopper.on_start(state)
        stopper.on_end_epoch(state)

        self.assertFalse(state[torchbearer.STOP_TRAINING])

        state[torchbearer.METRICS]['test_metric'] = 0.01
        stopper.on_end_epoch(state)

        self.assertFalse(state[torchbearer.STOP_TRAINING])

    def test_max_equal_should_stop(self):
        state = {
            torchbearer.EPOCH: 1,
            torchbearer.STOP_TRAINING: False,
            torchbearer.METRICS: {'test_metric': 0.001}
        }

        stopper = EarlyStopping(monitor='test_metric', mode='max')

        stopper.on_start(state)
        stopper.on_end_epoch(state)

        self.assertFalse(state[torchbearer.STOP_TRAINING])

        stopper.on_end_epoch(state)

        self.assertTrue(state[torchbearer.STOP_TRAINING])

    def test_in_equal_should_stop(self):
        state = {
            torchbearer.EPOCH: 1,
            torchbearer.STOP_TRAINING: False,
            torchbearer.METRICS: {'test_metric': 0.001}
        }

        stopper = EarlyStopping(monitor='test_metric', mode='min')

        stopper.on_start(state)
        stopper.on_end_epoch(state)

        self.assertFalse(state[torchbearer.STOP_TRAINING])

        stopper.on_end_epoch(state)

        self.assertTrue(state[torchbearer.STOP_TRAINING])

    def test_patience_should_stop(self):
        state = {
            torchbearer.EPOCH: 1,
            torchbearer.STOP_TRAINING: False,
            torchbearer.METRICS: {'test_metric': 0.001}
        }

        stopper = EarlyStopping(monitor='test_metric', patience=3)

        stopper.on_start(state)

        for i in range(3):
            stopper.on_end_epoch(state)
            self.assertFalse(state[torchbearer.STOP_TRAINING])

        stopper.on_end_epoch(state)
        self.assertTrue(state[torchbearer.STOP_TRAINING])

    def test_patience_should_continue(self):
        state = {
            torchbearer.EPOCH: 1,
            torchbearer.STOP_TRAINING: False,
            torchbearer.METRICS: {'test_metric': 0.001}
        }

        stopper = EarlyStopping(monitor='test_metric', patience=3)

        stopper.on_start(state)

        for i in range(3):
            stopper.on_end_epoch(state)
            self.assertFalse(state[torchbearer.STOP_TRAINING])

        state[torchbearer.METRICS]['test_metric'] = 0.0001
        stopper.on_end_epoch(state)
        self.assertFalse(state[torchbearer.STOP_TRAINING])

    def test_min_delta_should_continue(self):
        state = {
            torchbearer.EPOCH: 1,
            torchbearer.STOP_TRAINING: False,
            torchbearer.METRICS: {'test_metric': 0.001}
        }

        stopper = EarlyStopping(monitor='test_metric', mode='max', min_delta=0.1)

        stopper.on_start(state)
        stopper.on_end_epoch(state)

        self.assertFalse(state[torchbearer.STOP_TRAINING])

        state[torchbearer.METRICS]['test_metric'] = 0.102
        stopper.on_end_epoch(state)

        self.assertFalse(state[torchbearer.STOP_TRAINING])

    def test_min_delta_should_stop(self):
        state = {
            torchbearer.EPOCH: 1,
            torchbearer.STOP_TRAINING: False,
            torchbearer.METRICS: {'test_metric': 0.001}
        }

        stopper = EarlyStopping(monitor='test_metric', mode='max', min_delta=0.1)

        stopper.on_start(state)
        stopper.on_end_epoch(state)

        self.assertFalse(state[torchbearer.STOP_TRAINING])

        state[torchbearer.METRICS]['test_metric'] = 0.10
        stopper.on_end_epoch(state)

        self.assertTrue(state[torchbearer.STOP_TRAINING])

    def test_auto_should_be_min(self):
        state = {
            torchbearer.EPOCH: 1,
            torchbearer.STOP_TRAINING: False,
            torchbearer.METRICS: {'test_metric': 0.001}
        }

        stopper = EarlyStopping(monitor='test_metric')

        stopper.on_start(state)
        stopper.on_end_epoch(state)

        self.assertTrue(stopper.mode == 'min')

    def test_auto_should_be_max(self):
        state = {
            torchbearer.EPOCH: 1,
            torchbearer.STOP_TRAINING: False,
            torchbearer.METRICS: {'acc_metric': 0.001}
        }

        stopper = EarlyStopping(monitor='acc_metric')

        stopper.on_start(state)
        stopper.on_end_epoch(state)

        self.assertTrue(stopper.mode == 'max')

    def test_monitor_should_continue(self):
        state = {
            torchbearer.EPOCH: 1,
            torchbearer.STOP_TRAINING: False,
            torchbearer.METRICS: {'test_metric_1': 0.001, 'test_metric_2': 0.001}
        }

        stopper = EarlyStopping(monitor='test_metric_2', mode='max')

        stopper.on_start(state)
        stopper.on_end_epoch(state)

        self.assertFalse(state[torchbearer.STOP_TRAINING])

        state[torchbearer.METRICS]['test_metric_1'] = 0.0001
        state[torchbearer.METRICS]['test_metric_2'] = 0.01
        stopper.on_end_epoch(state)

        self.assertFalse(state[torchbearer.STOP_TRAINING])

    def test_monitor_should_stop(self):
        state = {
            torchbearer.EPOCH: 1,
            torchbearer.STOP_TRAINING: False,
            torchbearer.METRICS: {'test_metric_1': 0.001, 'test_metric_2': 0.001}
        }

        stopper = EarlyStopping(monitor='test_metric_2', mode='max')

        stopper.on_start(state)
        stopper.on_end_epoch(state)

        self.assertFalse(state[torchbearer.STOP_TRAINING])

        state[torchbearer.METRICS]['test_metric_1'] = 0.1
        state[torchbearer.METRICS]['test_metric_2'] = 0.0001
        stopper.on_end_epoch(state)

        self.assertTrue(state[torchbearer.STOP_TRAINING])

    def test_state_dict(self):
        stopper = EarlyStopping(monitor='test_metric_1')
        stopper.wait = 10
        stopper.best = 20
        stopper.stopped_epoch = 30
        state = stopper.state_dict()

        stopper = EarlyStopping(monitor='test_metric_1')
        self.assertNotEqual(stopper.wait, 10)

        stopper.load_state_dict(state)
        self.assertEqual(stopper.wait, 10)
        self.assertEqual(stopper.best, 20)
        self.assertEqual(stopper.stopped_epoch, 30)




import unittest

import torch.nn as nn
import torch.optim as optim

import torchbearer
from torchbearer.metrics import LR


class TestLR(unittest.TestCase):
    def test_simple(self):
        state = {torchbearer.OPTIMIZER: optim.SGD(nn.Linear(10, 10).parameters(), lr=0.01)}
        metric = LR()
        metric.reset(state)
        self.assertDictEqual(metric.process(state), {'lr': 0.01})
        self.assertDictEqual(metric.process_final(state), {'lr': 0.01})

    def test_groups(self):
        state = {
            torchbearer.OPTIMIZER: optim.SGD([
                {'params': nn.Linear(10, 10).parameters()},
                {'params': nn.Linear(10, 10).parameters(), 'lr': 1e-3}
            ], lr=1e-2, momentum=0.9)
        }
        metric = LR()
        metric.reset(state)
        self.assertListEqual(metric.process(state)['lr'], [1e-2, 1e-3])
        self.assertListEqual(metric.process_final(state)['lr'], [1e-2, 1e-3])
